#!/usr/bin/env python

import argparse
import csv
import logging
import subprocess
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from itertools import batched
from pathlib import Path
from shutil import rmtree

import requests

from .util import is_current
from .traversal import traverse
from .. import automaton_types
from ..log import configure_logging
from ..traversal import Traversal

logger = logging.getLogger("dwdsmor")

project_dir = Path(__file__).parent.parent.parent

xsl_dir = project_dir / "share"
grammar_dir = project_dir / "grammar"
lexicon_dir = project_dir / "lexicon"


def run(cmd, output_file=None, log_file=None, **args):
    "Run a subprocess, optionally redirecting stdout/stderr to files."
    logger.debug("! %s", cmd)

    args["capture_output"] = (
        args["capture_output"]
        if "capture_output" in args
        else (output_file is not None or log_file is not None)
    )

    try:
        result = subprocess.run(
            cmd,
            **args,
            check=True,
            text=True,
            encoding="utf-8",
        )

        if output_file is not None:
            output_file.write_text(result.stdout)
        if log_file is not None:
            log_file.write_text(result.stderr)

        return result
    except subprocess.CalledProcessError as e:
        if output_file is not None and e.stdout is not None:
            output_file.write_text(e.stdout)
        if log_file is not None and e.stderr is not None:
            log_file.write_text(e.stderr)
        raise


def git_rev(git_dir):
    return run(
        ["git", "rev-parse", "--short", "HEAD"], capture_output=True, cwd=git_dir
    ).stdout.strip()


project_git_rev = git_rev(project_dir)


saxon = [
    "java",
    "-Xmx8g",
    "-cp",
    "/usr/share/java/Saxon-HE.jar",
    "net.sf.saxon.Transform",
]


def saxon_with_source_file(xml, xsl, *args, **kwargs):
    "Run an XSLT transformation of an XML source file via Saxon-HE."
    return run(
        [
            *saxon,
            f"-s:{xml.as_posix()}",
            f"-xsl:{xsl.as_posix()}",
            *args,
        ],
        **kwargs,
    )


def saxon_with_initial_template(xsl, *args, **kwargs):
    "Run an XSLT transformation with an initial template via Saxon-HE."
    return run(
        [
            *saxon,
            "-it",
            f"-xsl:{xsl.as_posix()}",
            *args,
        ],
        **kwargs,
    )


def has_sources(edition_dir):
    "Checks whether the source directory of an edition contains XML files."
    src_dir = edition_dir / "wb"
    for _source in src_dir.rglob("*.xml"):
        return True
    return False


def build_lexicon(edition_dir, automata_dir, force=False):
    "Builds the lexicon for an automaton edition."

    aux_dir = edition_dir / "aux"
    src_dir = edition_dir / "wb"
    sources = (
        list(xsl_dir.glob("*.xsl"))
        + list(src_dir.rglob("*.xml"))
        + list(aux_dir.glob("*.xml"))
        + list(edition_dir.glob("exclude.xml"))
    )

    manifest_xml = automata_dir / "manifest.xml"
    manifest_log = automata_dir / "manifest.log"
    lexicon_xml = automata_dir / "lexicon.xml"
    lexicon_log = automata_dir / "lexicon.log"
    lex_fst = automata_dir / "lex.fst"
    lex_log = automata_dir / "lex.log"

    if not force and is_current(manifest_xml, sources):
        logger.debug("Skip building manifest 'manifest.xml'")
        return False

    if not force and is_current(lexicon_xml, sources):
        logger.debug("Skip building intermediate lexicon 'lexicon.xml'")
        return False

    if not force and is_current(lex_fst, sources):
        logger.debug("Skip building lexicon 'lex.fst'")
        return False

    logger.info("Building manifest 'manifest.xml'")
    exclude_file = edition_dir / "exclude.xml"
    xsl_params = [f"dwds-dir={src_dir.as_posix()}"]
    if aux_dir.is_dir():
        xsl_params.append(f"aux-dir={aux_dir.as_posix()}")
    if exclude_file.is_file():
        xsl_params.append(f"exclude-file={exclude_file.as_posix()}")
    saxon_with_initial_template(
        xsl_dir / "dwds2manifest.xsl",
        *xsl_params,
        cwd=edition_dir,
        output_file=manifest_xml,
        log_file=manifest_log,
    )

    logger.info("Building intermediate lexicon 'lexicon.xml'")
    saxon_with_initial_template(
        xsl_dir / "dwds2lexicon.xsl",
        f"manifest-file={manifest_xml.as_posix()}",
        cwd=edition_dir,
        output_file=lexicon_xml,
        log_file=lexicon_log,
    )

    logger.info("Building lexicon 'lex.fst'")
    lex = saxon_with_source_file(
        lexicon_xml,
        xsl_dir / "lexicon2dwdsmor.xsl",
        cwd=edition_dir,
        log_file=lex_log,
    )
    lex_fst.write_text("\n".join(sorted(set(lex.stdout.splitlines()))))
    return True


fst_modules = [fst_file for fst_file in grammar_dir.glob("module_*.fst")]
fst_macros = [fst_file for fst_file in grammar_dir.glob("macro_*.fst")]


def build_automaton(edition_dir, automata_dir, automaton_type, force=False):
    "Builds an automaton of the given type and edition."

    lexicon_src = automata_dir / "lex.fst"
    lexicon_symlink = grammar_dir / "lex.fst"
    automaton_src = grammar_dir / f"{automaton_type}.fst"
    automaton_a = automata_dir / f"{automaton_type}.a"
    automaton_ca = automata_dir / f"{automaton_type}.ca"
    automaton_compile_log = automata_dir / f"{automaton_type}.log"

    sources = [lexicon_src, automaton_src, *fst_modules, *fst_macros]
    if (
        not force
        and is_current(automaton_a, sources)
        and is_current(automaton_ca, sources)
    ):
        logger.debug("Skip building automaton '%s'", automaton_type)
        return False

    logger.info("Building automaton '%s'", automaton_type)
    try:
        automata_dir.mkdir(parents=True, exist_ok=True)
        if lexicon_symlink.is_file():
            lexicon_symlink.unlink()
        lexicon_symlink.symlink_to(lexicon_src)
        run(
            ["fst-compiler-utf8", automaton_src.as_posix(), automaton_a.as_posix()],
            cwd=grammar_dir,
            log_file=automaton_compile_log,
        )
        run(["fst-compact", automaton_a.as_posix(), automaton_ca.as_posix()])
    finally:
        if lexicon_symlink.is_symlink():
            lexicon_symlink.unlink()
    return True


def build_ppm_table(automata_dir, force=False):
    index_automaton_a = automata_dir / "index.a"
    if not index_automaton_a.is_file():
        return False

    ppm_table = automata_dir / "ppm.csv"
    if not force and is_current(ppm_table, (index_automaton_a,)):
        logger.debug("Skip building PPM table")
        return False

    lexdb_corpora = ("dibilit", "dwdsxl", "dtaxl", "wikipedia")

    with ThreadPoolExecutor(max_workers=len(lexdb_corpora)) as lexdb_threads:

        def lexdb_query(query, to_result):
            def qf(corpus, query=query):
                url = f"https://ddc.dwds.de/dstar/{corpus}/lexdb/export.perl"
                r = requests.post(url, data=(query | {"fmt": "json"}))
                r.raise_for_status()
                return r.json()

            futures = (lexdb_threads.submit(qf, c) for c in lexdb_corpora)
            for c, f in zip(lexdb_corpora, futures):
                yield (c, to_result(f.result()))

        lexdb_corpus_sizes = dict(
            lexdb_query(
                {"select": "sum(f)", "from": "lex"},
                lambda r: int(r.get("rows", [["0"]])[0][0]),
            )
        )

        def sql_quote_lemma(lemma):
            lemma = lemma.replace("'", "")
            return f"'{lemma}'"

        def get_ppms(lemmata):
            query = {
                "select": "l, sum(f)",
                "from": "lex",
                "where": f"l in ({','.join(map(sql_quote_lemma, lemmata))})",
                "groupby": "l",
                "limit": str(len(lemmata)),
            }

            ppms = {}
            for corpus, freqs in lexdb_query(query, lambda r: r.get("rows", [])):
                for lemma, n in freqs:
                    ppms[lemma] = max(
                        (int(n) * 1000000) / lexdb_corpus_sizes[corpus],
                        ppms.get(lemma, 0),
                    )
            for lemma in lemmata:
                yield lemma, ppms.get(lemma, 0)

        logger.info("Building PPM table")

        analyses = defaultdict(set)
        for surface, spec in traverse(index_automaton_a):
            t = Traversal.parse(spec)
            if any(
                (t.pidx, t.orthinfo, t.metainfo, t.charinfo, t.syninfo, t.ellipinfo)
            ):
                continue
            analyses[(surface, t.pos)].add(t.analysis)

        ambiguous_analyses = set()
        for _surface, surface_analyses in analyses.items():
            if len(surface_analyses) > 1:
                ambiguous_analyses |= surface_analyses

        with ppm_table.open("wt") as ppm_table_f:
            ppm_table_csv = csv.writer(ppm_table_f)
            for batch in batched(sorted(ambiguous_analyses), 100):
                for lemma, ppm in get_ppms(batch):
                    ppm_table_csv.writerow((lemma, str(ppm)))

        return True


arg_parser = argparse.ArgumentParser(description="Build DWDSmor.")
arg_parser.add_argument(
    "-c", "--clean", help="clean automata dir before build", action="store_true"
)
arg_parser.add_argument(
    "-d",
    "--automata-dir",
    help="directory where built automata are stored",
    type=Path,
    default=str(project_dir / "dwdsmor" / "automata"),
)
arg_parser.add_argument("-e", "--edition", help="edition to build", default="open")
arg_parser.add_argument(
    "-t",
    "--automaton-type",
    choices=automaton_types,
    help="automaton type to build (default: all)",
    action="append",
)
arg_parser.add_argument(
    "-f",
    "--force",
    help="force building (also current targets)",
    action="store_true",
)
arg_parser.add_argument(
    "-q", "--quiet", help="do not report progress", action="store_true"
)
args = arg_parser.parse_args()
configure_logging()

edition = args.edition
edition_dir = lexicon_dir / edition
build_automaton_types = args.automaton_type if args.automaton_type else automaton_types
assert edition_dir.is_dir()
assert has_sources(edition_dir), f"Edition {str(edition_dir)} without sources"

automata_dir = args.automata_dir.resolve()
if automata_dir.is_dir() and args.clean:
    logger.info("Cleaning build dir '%s'", automata_dir)
    rmtree(automata_dir)

logger.info("Building edition '%s' @ '%s'", edition, str(automata_dir))
automata_dir.mkdir(parents=True, exist_ok=True)

build_time = datetime.now(timezone.utc).isoformat()
edition_built = build_lexicon(edition_dir, automata_dir, force=args.force)
for automaton_type in build_automaton_types:
    edition_built = (
        build_automaton(edition_dir, automata_dir, automaton_type, force=args.force)
        or edition_built
    )
edition_built = build_ppm_table(automata_dir, force=args.force) or edition_built
if edition_built:
    (automata_dir / "BUILT").write_text(build_time)
    (automata_dir / "EDITION").write_text(edition)
    (automata_dir / "GIT_REV").write_text(project_git_rev)



#!/usr/bin/env python

import argparse
import csv
import sys
from collections import Counter, defaultdict, namedtuple
from functools import cache
from pathlib import Path
from typing import Dict
from xml.etree.ElementTree import parse

from tabulate import tabulate
from tqdm import tqdm

from .traversal import traverse
from .. import default_automata_dir
from ..traversal import Traversal
from ..log import configure_logging


# mapping between relevant DWDS and DWDSmor part-of-speech categories
pos_map = {
    "Adjektiv": {"ADJ"},
    "bestimmter Artikel": {"ART"},
    "Bruchzahlwort": {"FRAC"},
    "Demonstrativpronomen": {"DEM", "PROADV"},
    "Eigenname": {"NPROP"},
    "Indefinitpronomen": {"INDEF", "ART"},
    "Interrogativpronomen": {"WPRO"},
    "Kardinalzahlwort": {"CARD"},
    "Ordinalzahlwort": {"ORD"},
    "partizipiales Adjektiv": {"ADJ"},
    "Personalpronomen": {"PPRO"},
    "Possessivpronomen": {"POSS"},
    "Reflexivpronomen": {"PPRO"},
    "Relativpronomen": {"REL"},
    "reziprokes Pronomen": {"PPRO"},
    "Substantiv": {"NN", "NPROP"},
    "unbestimmter Artikel": {"ART"},
    "Verb": {"V"},
}


@cache
def qn(ln):
    return "{http://www.dwds.de/ns/1.0}" + ln


Entry = namedtuple(
    "Entry",
    ["file", "entry", "lemma", "lidx", "pos", "has_grammar"],
)


def parse_entries(wb_dir, xml_file):
    entries = []
    root = parse(xml_file).getroot()
    for entry, article in enumerate(root.findall(qn("Artikel"))):
        entry += 1
        if article.get("Status") not in {"Red-f", None}:
            continue
        for formspec in article.findall(qn("Formangabe")):
            for spelling in formspec.findall(qn("Schreibung")):
                if spelling.get("Typ") is not None:
                    continue
                lemma = spelling.text or ""
                lemma = lemma.strip()
                # ignore idioms and other syntactically complex units
                if " " in lemma:
                    continue
                lidx = spelling.get("hidx")
                lidx = f"IDX{lidx}" if lidx else ""
                for grammarspec in formspec.findall(qn("Grammatik")):
                    for pos in grammarspec.findall(qn("Wortklasse")):
                        pos = pos.text or ""
                        pos = pos.strip()
                        # ignore extraneous part-of-speech categories:
                        if pos not in pos_map:
                            continue

                        def prop(xpath, grammarspec=grammarspec):
                            for _ in grammarspec.findall(xpath):
                                return True
                            return False

                        def not_prop(xpath, grammarspec=grammarspec):
                            for _ in grammarspec.findall(xpath):
                                return False
                            return True

                        if pos in {"Eigenname", "Substantiv"}:
                            has_grammar = (
                                prop(qn("Genus")) and prop(qn("Genitiv"))
                            ) or prop(qn("Numeruspraeferenz") + "[.='nur im Plural']")
                        elif pos in {
                            "Personalpronomen",
                        }:
                            has_grammar = not_prop(
                                qn("Kasuspraeferenz") + "[.!='im Nominativ']"
                            ) and not_prop(qn("Einschraenkung"))
                        elif pos in {
                            "bestimmter Artikel",
                            "Demonstrativpronomen",
                            "Indefinitpronomen",
                            "Interrogativpronomen",
                            "Possessivpronomen",
                            "Relativpronomen",
                            "unbestimmter Artikel",
                        }:
                            has_grammar = (
                                prop(qn("Genus") + "[.='fem.']")
                                or prop(qn("indeklinabel"))
                            ) and not_prop(
                                qn("Kasuspraeferenz") + "[.!='im Nominativ']"
                            )
                        elif pos == "Verb":
                            has_grammar = (
                                prop(qn("Praesens"))
                                or prop(qn("Praeteritum"))
                                or prop(qn("Partizip_II"))
                                or prop(qn("Auxiliar"))
                            )
                        else:
                            has_grammar = True

                        entries.append(
                            Entry(
                                str(xml_file.relative_to(wb_dir)),
                                entry,
                                lemma,
                                lidx,
                                pos,
                                has_grammar,
                            )
                        )
    return sorted(set(entries))


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(
        prog=__package__,
        description="Benchmarks coverage of DWDSmor against the lexicon source.",
    )
    arg_parser.add_argument(
        "-e",
        "--edition",
        help="edition to run the benchmark against (default: open)",
        default="open",
    )
    arg_parser.add_argument(
        "-d", "--debug", help="print coverage data to stdout", action="store_true"
    )
    arg_parser.add_argument(
        "-q", "--quiet", help="do not report progress", action="store_true"
    )
    arg_parser.add_argument(
        "--limit",
        help="number of lexicon entries used for the benchmark (default: all)",
        type=int,
    )
    args = arg_parser.parse_args()
    configure_logging()

    project_dir = Path(".").resolve()
    edition_build_dir = project_dir / "build" / args.edition
    wb_dir = project_dir / "lexicon" / args.edition / "wb"
    wb_xml_files = tuple(sorted(wb_dir.rglob("*.xml")))

    limit = min(max(0, args.limit), len(wb_xml_files)) if args.limit else None
    if limit is not None:
        wb_xml_files = wb_xml_files[:limit]

    assert len(wb_xml_files) > 0, f"No lexicon articles found in {wb_dir}"

    traversals = defaultdict(list)
    for _surface, spec in traverse(default_automata_dir / "index.a"):
        t = Traversal.parse(spec)
        traversals[t.analysis].append(t)

    aggregated = not args.debug
    report = csv.writer(sys.stdout) if not aggregated else None
    report_header = False
    if report is None:
        if not args.quiet:
            wb_xml_files = tqdm(wb_xml_files, "Processing lexicon articles")

    generated_entries: Dict[str, Dict[bool, int]] = defaultdict(Counter)
    for wb_xml_file in wb_xml_files:
        for entry in parse_entries(wb_dir, wb_xml_file):
            pos_candidates = pos_map[entry.pos]
            generated = False
            for g_traversal in traversals.get(entry.lemma, []):
                if g_traversal.lidx == entry.lidx and g_traversal.pos in pos_candidates:
                    generated = True
                    break
            if entry.has_grammar:
                if aggregated:
                    # skip entries without appropriate grammatical specifications
                    generated_entries[entry.pos][generated] += 1
                if report is not None:
                    if not report_header:
                        report.writerow(
                            (
                                "File",
                                "Entry Number",
                                "Lemma",
                                "Lemma Index",
                                "POS",
                                "Generated",
                            )
                        )
                        report_header = True
                    result_cols = (
                        entry.file,
                        entry.entry,
                        entry.lemma,
                        entry.lidx,
                        entry.pos,
                        "1" if generated else "0",
                    )
                    report.writerow(result_cols)
    if aggregated:
        headers = [
            "POS",
            "Lemmas",
            "% Lemmas",
            "% Generated Lemmas",
        ]
        colalign = ["left", "right", "right", "right"]
        coverage = []

        def pct(v):
            return "{:.2%}".format(v)

        def count(v):
            return "{:,d}".format(v)

        total_entries = 0
        for pos_key in sorted(generated_entries.keys()):
            pos_generated = generated_entries[pos_key][True]
            pos_not_generated = generated_entries[pos_key][False]
            total_entries += pos_generated + pos_not_generated

        total_generated = 0
        for pos_key in sorted(generated_entries.keys()):
            pos_generated = generated_entries[pos_key][True]
            pos_not_generated = generated_entries[pos_key][False]
            pos_total = pos_generated + pos_not_generated
            coverage.append(
                [
                    pos_key,
                    count(pos_total),
                    pct(pos_total / total_entries),
                    pct(pos_generated / pos_total),
                ]
            )
            total_generated += pos_generated

        coverage.insert(
            0,
            [
                "Σ",
                count(total_entries),
                pct(1),
                pct(total_generated / total_entries),
            ],
        )

        print(tabulate(coverage, headers=headers, colalign=colalign, tablefmt="tsv"))



import argparse
from dataclasses import asdict, astuple, dataclass

from tabulate import tabulate
from tqdm import tqdm

import dwdsmor

from . import edition
from .log import configure_logging
from .tag import all_tags, boundary_tags, get_taggings, lexeme_tags
from .traversal import Traversal
from .util import inflected
from .version import __version__

headers = {
    "word": "Wordform",
    "spec": "Analysis",
    "analysis": "Lemma",
    "segmented": "Segmented Wordform",
    "weights": "Weights",
    "lidx": "Lemma Index",
    "pidx": "Paradigm Index",
    "processes": "Processes",
    "means": "Means",
    "pos": "POS",
    "category": "Subcategory",
    "auxiliary": "Auxiliary",
    "degree": "Degree",
    "person": "Person",
    "gender": "Gender",
    "case": "Case",
    "number": "Number",
    "inflection": "Inflection",
    "nonfinite": "Nonfinite",
    "function": "Function",
    "mood": "Mood",
    "tense": "Tense",
    "metainfo": "Metalinguistic",
    "orthinfo": "Orthography",
    "syninfo": "Syntax",
    "ellipinfo": "Ellipsis",
    "charinfo": "Characters",
}


@dataclass
class Result(Traversal):
    word: str | None = None

    @staticmethod
    def from_traversal(traversal: Traversal, word):
        return Result(**asdict(traversal), word=word)

    @staticmethod
    def from_spec(spec: str, word: str, idx_to_int=False):
        return Result(**asdict(Traversal.parse(spec, idx_to_int=idx_to_int)), word=word)


def generate(analyzer, generator, word, quiet=True, idx_to_int=False):
    lexeme_specs = set()
    for traversal in analyzer.analyze(word):
        traversal = traversal.reparse(visible_boundaries=boundary_tags)
        lexeme_spec = traversal.analysis
        for lexeme_tag in lexeme_tags:
            tag = getattr(traversal, lexeme_tag)
            if tag is not None:
                lexeme_spec += f"<{tag}>"
        lexeme_specs.add((traversal.pos, lexeme_spec))
    for pos, lexeme_spec in sorted(lexeme_specs):
        results = []
        taggings = get_taggings(pos)
        if not quiet:
            taggings = tqdm(
                taggings,
                desc=f"Inflection patterns ({pos})",
                total=len(taggings),
            )
        for tagging in taggings:
            spec = lexeme_spec + "".join(f"<{tag}>" for tag in tagging)
            for generated in generator.generate(spec):
                infl_form = inflected(spec, generated.spec)
                results.append(Result.from_spec(spec, infl_form, idx_to_int=idx_to_int))
        for result in sorted(results, key=lambda r: r.spec):
            yield result


def main():
    arg_parser = argparse.ArgumentParser(description="Traverse DWDSmor automata.")
    arg_parser.add_argument(
        "-t", "--automaton-type", choices=dwdsmor.automaton_types, help="automaton type"
    )
    arg_parser.add_argument("-g", "--generate", help="generate", action="store_true")
    arg_parser.add_argument(
        "-q", "--quiet", help="do not report progress", action="store_true"
    )
    arg_parser.add_argument(
        "-v", "--version", help="show version and exit", action="store_true"
    )
    arg_parser.add_argument("words", nargs="*")
    args = arg_parser.parse_args()
    configure_logging()

    if args.version:
        print(f"DWDSmor v{__version__} (edition: {edition})")
        return 0

    results = []

    automaton = "index" if args.generate else "lemma"
    automaton = args.automaton_type if args.automaton_type else automaton

    visible_boundaries = "+" if automaton == "root" or automaton == "root2" else ""
    boundary_tag = " + " if automaton == "root" or automaton == "root2" else None
    idx_to_int = True if automaton == "index" else False
    join_tags = True

    if args.generate:
        analyzer = dwdsmor.analyzer(automaton)
        generator = dwdsmor.generator(automaton)
        for word in args.words:
            for result in generate(analyzer, generator, word, args.quiet, idx_to_int):
                results.append(result)
    else:
        analyzer = dwdsmor.analyzer(automaton)
        for word in args.words:
            local_results = []
            max_boundaries = {}
            for traversal in analyzer.analyze(
                word,
                visible_boundaries=visible_boundaries,
                boundary_tag=boundary_tag,
                idx_to_int=idx_to_int,
                join_tags=join_tags,
            ):
                with_boundaries = traversal.reparse(
                    visible_boundaries=boundary_tags, boundary_tag="|"
                )
                boundaries = with_boundaries.analysis.count("|")
                # group by traversal props, leaving out discriminating
                # spec, which expresses boundaries
                group_by = astuple(traversal)[1:]
                if max_boundaries.get(group_by, -1) < boundaries:
                    max_boundaries[group_by] = boundaries
                    local_results.append(Result.from_traversal(traversal, word))
            local_results = sorted(local_results, key=lambda r: r.spec)
            results.extend(local_results)

    if not results:
        return 1

    def has_value(tag_type, results=results):
        for result in results:
            if getattr(result, tag_type) is not None:
                return True
        return False

    cols = [
        "word",
        "analysis",
        "spec",
        *(tt for tt in all_tags if has_value(tt)),
    ]

    print(
        tabulate(
            [[getattr(r, c) for c in cols] for r in results],
            headers=[headers.get(c, c) for c in cols],
            tablefmt="tsv",
        )
    )

    return 0



import csv

import dwdsmor

from .conftest import if_dwds_available, test_dir


@if_dwds_available
def test_numeral():
    analyzer = dwdsmor.analyzer()
    test_cases_file = test_dir / "numeral_test_cases.csv"
    with test_cases_file.open("rt") as f:
        for n, (pos, lemma) in enumerate(csv.reader(f)):
            if n == 0:
                continue
            assert (pos, lemma) in {
                (a.pos, a.analysis) for a in analyzer.analyze(lemma)
            }



#!/usr/bin/env python3
# paradigm.py -- generate paradigms with DWDSmor
# Andreas Nolda 2026-01-08

import argparse
import csv
import json
import sys
from collections import defaultdict, namedtuple
from itertools import filterfalse, product
from pathlib import Path

import yaml
from blessings import Terminal

import dwdsmor
from dwdsmor import tag

from .analysis import analyze_word, format_path, generate_words, seg_lemma


progname = "dwdsmor-paradigm"

version = 19.2


IDX = range(1, 9)


POS = ("NN", "NPROP", "ADJ", "CARD", "ORD", "FRAC", "ART",
       "DEM", "INDEF", "POSS", "PPRO", "REL", "WPRO", "V")

SUBCATS = ("Def", "Indef", "Neg", "Pers", "Refl", "Rec")

DEGREES = ("Pos", "Comp", "Sup")

PERSONS = ("1", "2", "3")

GENDERS = ("Masc", "Neut", "Fem", "UnmGend")

CASES = ("Nom", "Acc", "Dat", "Gen", "UnmCase")

NUMBERS = ("Sg", "Pl", "UnmNum")

INFLECTIONS = ("St", "Wk", "UnmInfl")

FUNCTIONS = ("Attr", "Subst", "Attr/Subst", "Pred/Adv", "UnmFunc")

MOODS = ("Ind", "Subj")

TENSES = ("Pres", "Perf", "Past", "PastPerf", "Fut", "FutPerf")

AUXILIARIES = ("haben", "sein")

SYNINFO = ("MEAS",)


NONST = "NonSt"

OLD = "Old"

OLDORTH = "OLDORTH"

CH = "CH"


TAG_NONST = "ugs."

TAG_OLD = "va."

TAG_OLDORTH = "ung."

TAG_CH = "CH"


SEG_HABEN = "hab<~>en"

SEG_SEIN = "sei<~>n"

SEG_WERDEN = "werd<~>en"


LIDX_HABEN = None

LIDX_SEIN = 3

LIDX_WERDEN = None


PIDX_HABEN = None

PIDX_SEIN = None

PIDX_WERDEN = 1


LEXCAT = ["pos",
          "subcat",
          "auxiliary",
          "person",
          "gender"]

PARCAT = ["degree",
          "person",
          "gender",
          "case",
          "number",
          "inflection",
          "nonfinite",
          "function",
          "mood",
          "tense",
          "syninfo"]


LABEL_MAP = {"lemma": "Lemma",
             "lidx": "Lemma Index",
             "pidx": "Paradigm Index",
             "categories": "Categories",
             "pos": "POS",
             "subcat": "Subcategory",
             "auxiliary": "Auxiliary",
             "person": "Person",
             "gender": "Gender",
             "paradigm": {"categories": "Paradigm Categories",
                          "degree": "Degree",
                          "person": "Person",
                          "gender": "Gender",
                          "case": "Case",
                          "number": "Number",
                          "inflection": "Inflection",
                          "nonfinite": "Nonfinite",
                          "function": "Function",
                          "mood": "Mood",
                          "tense": "Tense",
                          "syninfo": "Syntax",
                          "forms": "Paradigm Forms"}}


Lexcat = namedtuple("Lexcat", LEXCAT, defaults=[None] * len(LEXCAT[1:]))

Parcat = namedtuple("Parcat", PARCAT, defaults=[None] * len(PARCAT))


Formspec = namedtuple("Formspec", ["lidx", "pidx",
                                   "lexcat", "parcat"])

Lemmaspec = namedtuple("Lemmaspec", ["lidx", "pidx", "seg_lemma",
                                     "pos", "subcat", "person", "gender"])


def filter_categorisations(categorisations, pos):
    if pos in ["ADJ", "ART", "CARD", "DEM", "FRAC", "INDEF", "NN",
               "NPROP", "ORD", "POSS", "PPRO", "REL", "WPRO"]:
        # Nom, Acc, Dat, and Gen do not co-occur with UnmNum.
        categorisations = filterfalse(lambda cat: "Nom" in cat and "UnmNum" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Acc" in cat and "UnmNum" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Dat" in cat and "UnmNum" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Gen" in cat and "UnmNum" in cat,
                                      categorisations)
        # Nom, Acc, Dat, and Gen do not co-occur with UnmFunc.
        categorisations = filterfalse(lambda cat: "Nom" in cat and "UnmFunc" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Acc" in cat and "UnmFunc" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Dat" in cat and "UnmFunc" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Gen" in cat and "UnmFunc" in cat,
                                      categorisations)
        # Sg and Pl do not co-occur with UnmFunc.
        categorisations = filterfalse(lambda cat: "Sg" in cat and "UnmFunc" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Pl" in cat and "UnmFunc" in cat,
                                      categorisations)
    if pos in ["ADJ", "ART", "CARD", "DEM", "FRAC", "INDEF", "NN",
               "ORD", "POSS", "REL", "WPRO"]:
        # UnmGend does not co-occur with Sg.
        categorisations = filterfalse(lambda cat: "UnmGend" in cat and "Sg" in cat,
                                      categorisations)
        # Masc, Neut, and Fem do not co-occur with Pl.
        categorisations = filterfalse(lambda cat: "Masc" in cat and "Pl" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Neut" in cat and "Pl" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Fem" in cat and "Pl" in cat,
                                      categorisations)
        # Masc, Neut, and Fem do not co-occur with UnmCase.
        categorisations = filterfalse(lambda cat: "Masc" in cat and "UnmCase" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Neut" in cat and "UnmCase" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Fem" in cat and "UnmCase" in cat,
                                      categorisations)
        # Masc, Neut, and Fem do not co-occur with UnmNum.
        categorisations = filterfalse(lambda cat: "Masc" in cat and "UnmNum" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Neut" in cat and "UnmNum" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Fem" in cat and "UnmNum" in cat,
                                      categorisations)
        # Masc, Neut, and Fem do not co-occur with UnmFunc.
        categorisations = filterfalse(lambda cat: "Masc" in cat and "UnmFunc" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Neut" in cat and "UnmFunc" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Fem" in cat and "UnmFunc" in cat,
                                      categorisations)
        # Dat and Gen do not co-occur with UnmInfl.
        categorisations = filterfalse(lambda cat: "Dat" in cat and "UnmInfl" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Gen" in cat and "UnmInfl" in cat,
                                      categorisations)
        # St and Wk do not co-occur with UnmCase.
        categorisations = filterfalse(lambda cat: "St" in cat and "UnmCase" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Wk" in cat and "UnmCase" in cat,
                                      categorisations)
        # St and Wk do not co-occur with UnmNum.
        categorisations = filterfalse(lambda cat: "St" in cat and "UnmNum" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Wk" in cat and "UnmNum" in cat,
                                      categorisations)
        # St and Wk do not co-occur with UnmFunc.
        categorisations = filterfalse(lambda cat: "St" in cat and "UnmFunc" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Wk" in cat and "UnmFunc" in cat,
                                      categorisations)
    if pos in ["ADJ", "DEM", "NN", "ORD", "REL", "WPRO"]:
        # Sg and Pl do not co-occur with UnmInfl.
        categorisations = filterfalse(lambda cat: "Sg" in cat and "UnmInfl" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Pl" in cat and "UnmInfl" in cat,
                                      categorisations)
    if pos in ["ART", "CARD", "DEM", "INDEF", "POSS", "REL", "WPRO"]:
        # There are no categories Attr/Subst or Pred/Adv.
        categorisations = filterfalse(lambda cat: "Attr/Subst" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Pred/Adv" in cat,
                                      categorisations)
    if pos in ["ADJ", "ORD"]:
        # There is no category Subst.
        categorisations = filterfalse(lambda cat: "Subst" in cat,
                                      categorisations)
        # Pred/Adv does not co-occur with any category.
        categorisations = filterfalse(lambda cat: "Pred/Adv" in cat and len(cat) > 1,
                                      categorisations)
    if pos == "NN":
        # Sg and Pl do not co-occur with MEAS.
        categorisations = filterfalse(lambda cat: "Sg" in cat and "MEAS" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Pl" in cat and "MEAS" in cat,
                                      categorisations)
        # Nom, Acc, Dat, and Gen do not co-occur with MEAS.
        categorisations = filterfalse(lambda cat: "Nom" in cat and "MEAS" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Acc" in cat and "MEAS" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Dat" in cat and "MEAS" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Gen" in cat and "MEAS" in cat,
                                      categorisations)
    if pos == "V":
        # Ind and Subj do not co-occur with UnmNum.
        categorisations = filterfalse(lambda cat: "Ind" in cat and "UnmNum" in cat,
                                      categorisations)
        categorisations = filterfalse(lambda cat: "Subj" in cat and "UnmNum" in cat,
                                      categorisations)
    return categorisations


def format_orthinfo(orthinfo):
    return f"<{orthinfo}>" if orthinfo else ""


def format_lidx(lidx):
    return f"<IDX{lidx}>" if lidx else ""


def format_pidx(pidx):
    return f"<PAR{pidx}>" if pidx else ""


def format_pos(pos):
    return f"<{pos}>"


def format_categorisation(categorisation):
    return "".join([f"<{cat}>" for cat in categorisation])


def format_inf_cl(form):
    return "zu" + " " + form


def format_prev_verb_form(form, prev):
    return form[len(prev):] + " " + prev


def format_double_prev_verb_form(form, prev, prev2):
    return form[len(prev2 + prev):] + " " + prev2 + " " + prev


def format_tags(tags):
    return "(" + ", ".join(tags) + ")"


def generate_forms(generator, lidx, pidx, seg,
                   pos, categorisation, orthinfo=""):
    forms = generate_words(generator, format_orthinfo(orthinfo) + seg
                                                                + format_lidx(lidx)
                                                                + format_pidx(pidx)
                                                                + format_pos(pos)
                                                                + format_categorisation(categorisation))
    return forms


def add_forms(formdict, lidx, pidx,
              lexcat, parcat, forms):
    if forms:
        formdict[Formspec(lidx, pidx,
                          lexcat, parcat)].extend(set(forms))


def add_special_forms(formdict, lidx, pidx,
                      lexcat, parcat, forms,
                      tags):
    if forms:
        formatted_forms = [form + " " + format_tags(tags)
                           for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, formatted_forms)


def add_sup_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms):
    if forms:
        complex_forms = ["am " + form
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_sup_forms(formdict, lidx, pidx,
                          lexcat, parcat, forms,
                          tags):
    if forms:
        complex_forms = ["am " + form + " " + format_tags(tags)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_perf_forms(formdict, lidx, pidx,
                   lexcat, parcat, forms, parts):
    if forms:
        complex_forms = [form + " " + part
                         for form in forms
                         for part in parts]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_perf_forms(formdict, lidx, pidx,
                           lexcat, parcat, forms, parts,
                           tags):
    if forms:
        complex_forms = [form + " " + part + " " + format_tags(tags)
                         for form in forms
                         for part in parts]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_fut_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms, infs):
    if forms:
        complex_forms = [form + " " + inf
                         for form in forms
                         for inf in infs]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_fut_forms(formdict, lidx, pidx,
                          lexcat, parcat, forms, infs,
                          tags):
    if forms:
        complex_forms = [form + " " + inf + " " + format_tags(tags)
                         for form in forms
                         for inf in infs]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_futperf_forms(formdict, lidx, pidx,
                      lexcat, parcat, forms, infs, parts):
    if forms:
        complex_forms = [form + " " + part + " " + inf
                         for form in forms
                         for part in parts
                         for inf in infs]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_futperf_forms(formdict, lidx, pidx,
                              lexcat, parcat, forms, infs, parts,
                              tags):
    if forms:
        complex_forms = [form + " " + part + " " + inf + " " + format_tags(tags)
                         for form in forms
                         for part in parts
                         for inf in infs]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_nonfinite_perf_forms(formdict, lidx, pidx,
                             lexcat, parcat, forms, parts):
    if forms:
        complex_forms = [part + " " + form
                         for part in parts
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_nonfinite_perf_forms(formdict, lidx, pidx,
                                     lexcat, parcat, forms, parts,
                                     tags):
    if forms:
        complex_forms = [part + " " + form + " " + format_tags(tags)
                         for part in parts
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_prev_verb_forms(formdict, lidx, pidx,
                        lexcat, parcat, forms, prev):
    if forms:
        complex_forms = [format_prev_verb_form(form, prev)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_prev_verb_forms(formdict, lidx, pidx,
                                lexcat, parcat, forms, prev,
                                tags):
    if forms:
        complex_forms = [format_prev_verb_form(form, prev) + " " + format_tags(tags)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_double_prev_verb_forms(formdict, lidx, pidx,
                               lexcat, parcat, forms, prev, prev2):
    if forms:
        complex_forms = [format_double_prev_verb_form(form, prev, prev2)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                       lexcat, parcat, forms, prev, prev2,
                                       tags):
    if forms:
        complex_forms = [format_double_prev_verb_form(form, prev, prev2) + " " + format_tags(tags)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_imp_prev_verb_forms(formdict, lidx, pidx,
                            lexcat, parcat, forms, prev):
    if forms:
        complex_forms = [format_prev_verb_form(form, prev)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_imp_prev_verb_forms(formdict, lidx, pidx,
                                    lexcat, parcat, forms, prev,
                                    tags):
    if forms:
        complex_forms = [format_prev_verb_form(form, prev) + " " + format_tags(tags)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_imp_double_prev_verb_forms(formdict, lidx, pidx,
                                   lexcat, parcat, forms, prev, prev2):
    if forms:
        complex_forms = [format_double_prev_verb_form(form, prev, prev2)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def add_special_imp_double_prev_verb_forms(formdict, lidx, pidx,
                                           lexcat, parcat, forms, prev, prev2,
                                           tags):
    if forms:
        complex_forms = [format_double_prev_verb_form(form, prev, prev2) + " " + format_tags(tags)
                         for form in forms]
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, complex_forms)


def get_noun_formdict(generator, lidx, pidx, seg,
                      pos, gender, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    lexcat = Lexcat(pos=pos,
                    gender=gender)
    # nominalised adjectives
    categorisations = product(NUMBERS, CASES, INFLECTIONS)
    categorisations = filter_categorisations(categorisations, pos)
    for number, case, inflection in categorisations:
        parcat = Parcat(case=case,
                        number=number,
                        inflection=inflection)
        categorisation = [gender,
                          case,
                          number,
                          inflection]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            # no such forms
            pass
        if old:
            # no such forms
            pass
        if oldorth:
            oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                           pos, categorisation, OLDORTH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, oldorth_forms,
                              [TAG_OLDORTH])
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
        if ch:
            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                      pos, categorisation, CH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, ch_forms,
                              [TAG_CH])
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass

    # measure nouns
    categorisations = product(NUMBERS, CASES, SYNINFO)
    categorisations = filter_categorisations(categorisations, pos)
    for number, case, syninfo in categorisations:
        parcat = Parcat(case=case,
                        number=number,
                        syninfo=syninfo)
        categorisation = [gender,
                          case,
                          number,
                          syninfo]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            # no such forms
            pass
        if old:
            # no such forms
            pass
        if oldorth:
            oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                           pos, categorisation, OLDORTH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, oldorth_forms,
                              [TAG_OLDORTH])
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
        if ch:
            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                      pos, categorisation, CH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, ch_forms,
                              [TAG_CH])
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass

    # other nouns
    categorisations = product(NUMBERS, CASES)
    categorisations = filter_categorisations(categorisations, pos)
    for number, case in categorisations:
        parcat = Parcat(case=case,
                        number=number)
        categorisation = [gender,
                          case,
                          number]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                         pos, categorisation + [NONST])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, nonst_forms,
                              [TAG_NONST])
        if old:
            old_forms = generate_forms(generator, lidx, pidx, seg,
                                       pos, categorisation + [OLD])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, old_forms,
                              [TAG_OLD])
        if oldorth:
            oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                           pos, categorisation, OLDORTH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, oldorth_forms,
                              [TAG_OLDORTH])
            if nonst:
                nonst_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                     pos, categorisation + [NONST], OLDORTH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, nonst_oldorth_forms,
                                  [TAG_NONST, TAG_OLDORTH])
            if old:
                old_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                   pos, categorisation + [OLD], OLDORTH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, old_oldorth_forms,
                                  [TAG_OLD, TAG_OLDORTH])
        if ch:
            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                      pos, categorisation, CH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, ch_forms,
                              [TAG_CH])
            if nonst:
                nonst_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                pos, categorisation + [NONST], CH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, nonst_ch_forms,
                                  [TAG_NONST, TAG_CH])
            if old:
                old_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                              pos, categorisation + [OLD], CH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, old_ch_forms,
                                  [TAG_OLD, TAG_CH])
    return formdict


def get_adjective_formdict(generator, lidx, pidx, seg,
                           pos, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    lexcat = Lexcat(pos=pos)
    for degree in DEGREES:
        # non-attributive forms
        parcat = Parcat(degree=degree,
                        function="Pred/Adv")
        categorisation = [degree,
                          "Pred/Adv"]
        nonattributive_forms = generate_forms(generator, lidx, pidx, seg,
                                              pos, categorisation)
        if degree == "Sup":
            add_sup_forms(formdict, lidx, pidx,
                          lexcat, parcat, nonattributive_forms)
        else:
            add_forms(formdict, lidx, pidx,
                      lexcat, parcat, nonattributive_forms)
        if nonst:
            # no such forms
            pass
        if old:
            # no such forms
            pass
        if oldorth:
            oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                           pos, categorisation, OLDORTH)
            if degree == "Sup":
                add_special_sup_forms(formdict, lidx, pidx,
                                      lexcat, parcat, oldorth_forms,
                                      [TAG_OLDORTH])
            else:
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, oldorth_forms,
                                  [TAG_OLDORTH])
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
        if ch:
            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                      pos, categorisation, CH)
            if degree == "Sup":
                add_special_sup_forms(formdict, lidx, pidx,
                                      lexcat, parcat, ch_forms,
                                      [TAG_CH])
            else:
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, ch_forms,
                                  [TAG_CH])
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass

    for degree in DEGREES:
        # forms inflected for degree, gender, case, number, and inflectional strength
        categorisations = product(GENDERS, NUMBERS, CASES, INFLECTIONS, FUNCTIONS)
        categorisations = filter_categorisations(categorisations, pos)
        for gender, number, case, inflection, function in categorisations:
            parcat = Parcat(degree=degree,
                            function=function,
                            gender=gender,
                            case=case,
                            number=number,
                            inflection=inflection)
            categorisation = [degree,
                              function,
                              gender,
                              case,
                              number,
                              inflection]
            forms = generate_forms(generator, lidx, pidx, seg,
                                   pos, categorisation)
            add_forms(formdict, lidx, pidx,
                      lexcat, parcat, forms)
            if nonst:
                # no such forms
                pass
            if old:
                old_forms = generate_forms(generator, lidx, pidx, seg,
                                           pos, categorisation + [OLD])
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, old_forms,
                                  [TAG_OLD])
            if oldorth:
                oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                               pos, categorisation, OLDORTH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, oldorth_forms,
                                  [TAG_OLDORTH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    old_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                       pos, categorisation + [OLD], OLDORTH)
                    add_special_forms(formdict, lidx, pidx,
                                      lexcat, parcat, old_oldorth_forms,
                                      [TAG_OLD, TAG_OLDORTH])
            if ch:
                ch_forms = generate_forms(generator, lidx, pidx, seg,
                                          pos, categorisation, CH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, ch_forms,
                                  [TAG_CH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    old_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                  pos, categorisation + [OLD], CH)
                    add_special_forms(formdict, lidx, pidx,
                                      lexcat, parcat, old_ch_forms,
                                      [TAG_OLD, TAG_CH])
    return formdict


def get_article_formdict(generator, lidx, pidx, seg,
                         pos, subcat, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    lexcat = Lexcat(pos=pos,
                    subcat=subcat)
    categorisations = product(GENDERS, NUMBERS, CASES, INFLECTIONS, FUNCTIONS)
    categorisations = filter_categorisations(categorisations, pos)
    for gender, number, case, inflection, function in categorisations:
        parcat = Parcat(function=function,
                        gender=gender,
                        case=case,
                        number=number,
                        inflection=inflection)
        categorisation = [subcat,
                          function,
                          gender,
                          case,
                          number,
                          inflection]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                         pos, categorisation + [NONST])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, nonst_forms,
                              [TAG_NONST])
        if old:
            # no such forms
            pass
        if oldorth:
            # no such forms
            pass
        if ch:
            # no such forms
            pass
    return formdict


def get_cardinal_formdict(generator, lidx, pidx, seg,
                          pos, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    lexcat = Lexcat(pos=pos)
    # forms inflected for function, gender, case, number, and inflectional strength
    categorisations = product(GENDERS, NUMBERS, CASES, INFLECTIONS, FUNCTIONS)
    categorisations = filter_categorisations(categorisations, pos)
    for gender, number, case, inflection, function in categorisations:
        parcat = Parcat(function=function,
                        gender=gender,
                        case=case,
                        number=number,
                        inflection=inflection)
        categorisation = [function,
                          gender,
                          case,
                          number,
                          inflection]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                         pos, categorisation + [NONST])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, nonst_forms,
                              [TAG_NONST])
        if old:
            # no such forms
            pass
        if oldorth:
            # no such forms
            pass
        if ch:
            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                      pos, categorisation, CH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, ch_forms,
                              [TAG_CH])
            if nonst:
                nonst_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                pos, categorisation + [NONST], CH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, nonst_ch_forms,
                                  [TAG_NONST, TAG_CH])
            if old:
                # no such forms
                pass
    return formdict


def get_ordinal_formdict(generator, lidx, pidx, seg,
                         pos, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    lexcat = Lexcat(pos=pos)
    categorisations = product(GENDERS, NUMBERS, CASES, INFLECTIONS, FUNCTIONS)
    categorisations = filter_categorisations(categorisations, pos)
    for gender, number, case, inflection, function in categorisations:
        parcat = Parcat(function=function,
                        gender=gender,
                        case=case,
                        number=number,
                        inflection=inflection)
        categorisation = [function,
                          gender,
                          case,
                          number,
                          inflection]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            # no such forms
            pass
        if old:
            # no such forms
            pass
        if oldorth:
            # no such forms
            pass
        if ch:
            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                      pos, categorisation, CH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, ch_forms,
                              [TAG_CH])
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
    return formdict


def get_fraction_formdict(generator, lidx, pidx, seg,
                          pos, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    lexcat = Lexcat(pos=pos)
    # forms inflected for function, gender, case, number, and inflectional strength
    categorisations = product(GENDERS, NUMBERS, CASES, INFLECTIONS, FUNCTIONS)
    categorisations = filter_categorisations(categorisations, pos)
    for gender, number, case, inflection, function in categorisations:
        parcat = Parcat(function=function,
                        gender=gender,
                        case=case,
                        number=number,
                        inflection=inflection)
        categorisation = [function,
                          gender,
                          case,
                          number,
                          inflection]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            # no such forms
            pass
        if old:
            # no such forms
            pass
        if oldorth:
            # no such forms
            pass
        if ch:
            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                      pos, categorisation, CH)
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, ch_forms,
                              [TAG_CH])
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
    return formdict


def get_adjectival_pronoun_formdict(generator, lidx, pidx, seg,
                                    pos, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    lexcat = Lexcat(pos=pos)
    # forms inflected for function, gender, case, number, and inflectional strength
    categorisations = product(GENDERS, NUMBERS, CASES, INFLECTIONS, FUNCTIONS)
    categorisations = filter_categorisations(categorisations, pos)
    for gender, number, case, inflection, function in categorisations:
        parcat = Parcat(function=function,
                        gender=gender,
                        case=case,
                        number=number,
                        inflection=inflection)
        categorisation = [function,
                          gender,
                          case,
                          number,
                          inflection]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                         pos, categorisation + [NONST])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, nonst_forms,
                              [TAG_NONST])
        if old:
            # no such forms
            pass
        if oldorth:
            # no such forms
            pass
        if ch:
            # no such forms
            pass
    return formdict


def get_substantival_pronoun_formdict(generator, lidx, pidx, seg,
                                      pos, subcat, person, gender, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    categorisations = product(NUMBERS, CASES)
    categorisations = filter_categorisations(categorisations, pos)
    for number, case in categorisations:
        if gender:
            lexcat = Lexcat(pos=pos,
                            subcat=subcat,
                            person=person,
                            gender=gender)
            parcat = Parcat(case=case,
                            number=number)
            categorisation = [subcat,
                              person,
                              gender,
                              case,
                              number]
        else:
            lexcat = Lexcat(pos=pos,
                            subcat=subcat,
                            person=person)
            parcat = Parcat(case=case,
                            number=number)
            categorisation = [subcat,
                              person,
                              case,
                              number]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                         pos, categorisation + [NONST])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, nonst_forms,
                              [TAG_NONST])
        if old:
            old_forms = generate_forms(generator, lidx, pidx, seg,
                                       pos, categorisation + [OLD])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, old_forms,
                              [TAG_OLD])
        if oldorth:
            # no such forms
            pass
        if ch:
            # no such forms
            pass
    return formdict


def get_other_pronoun_formdict(generator, lidx, pidx, seg,
                               pos, gender, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    # fixed gender
    lexcat = Lexcat(pos=pos,
                    gender=gender)
    # forms with fixed gender and inflected for case and number
    categorisations = product(NUMBERS, CASES)
    categorisations = filter_categorisations(categorisations, pos)
    for number, case in categorisations:
        parcat = Parcat(case=case,
                        number=number)
        categorisation = [gender,
                          case,
                          number]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                         pos, categorisation + [NONST])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, nonst_forms,
                              [TAG_NONST])
        if old:
            old_forms = generate_forms(generator, lidx, pidx, seg,
                                       pos, categorisation + [OLD])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, old_forms,
                              [TAG_OLD])
        if oldorth:
            # no such forms
            pass
        if ch:
            # no such forms
            pass

    # no fixed gender
    lexcat = Lexcat(pos=pos)
    # forms inflected for function, gender, case, number, and inflectional strength
    categorisations = product(GENDERS, NUMBERS, CASES, INFLECTIONS, FUNCTIONS)
    categorisations = filter_categorisations(categorisations, pos)
    for gender, number, case, inflection, function in categorisations:
        parcat = Parcat(function=function,
                        gender=gender,
                        case=case,
                        number=number,
                        inflection=inflection)
        categorisation = [function,
                          gender,
                          case,
                          number,
                          inflection]
        forms = generate_forms(generator, lidx, pidx, seg,
                               pos, categorisation)
        add_forms(formdict, lidx, pidx,
                  lexcat, parcat, forms)
        if nonst:
            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                         pos, categorisation + [NONST])
            add_special_forms(formdict, lidx, pidx,
                              lexcat, parcat, nonst_forms,
                              [TAG_NONST])
        if old:
            # no such forms
            pass
        if oldorth:
            # no such forms
            pass
        if ch:
            # no such forms
            pass
    return formdict


def get_verb_formdict(generator, lidx, pidx, seg,
                      pos, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    word_list = seg.split(f"<{tag.prev_boundary}>")
    word_count = len(word_list)
    prev = word_list[1] if word_count == 3 else word_list[0] if word_count == 2 else ""
    prev2 = word_list[0] if word_count == 3 else ""
    for auxiliary in AUXILIARIES:
        categorisation = ["Part",
                          "Perf",
                          auxiliary]
        part_perf_forms = generate_forms(generator, lidx, pidx, seg,
                                         pos, categorisation)
        if oldorth:
            oldorth_part_perf_forms = generate_forms(generator, lidx, pidx, seg,
                                                     pos, categorisation, OLDORTH)
        if ch:
            ch_part_perf_forms = generate_forms(generator, lidx, pidx, seg,
                                                pos, categorisation, CH)
        # test whether the past participle actually selects the auxiliary
        if part_perf_forms:
            lexcat = Lexcat(pos=pos,
                            auxiliary=auxiliary)
            # infinitives
            parcat = Parcat(nonfinite="Inf",
                            function="NonCl",
                            tense="Pres")
            categorisation = ["Inf",
                              "NonCl"]
            inf_forms = generate_forms(generator, lidx, pidx, seg,
                                       pos, categorisation)
            add_forms(formdict, lidx, pidx,
                      lexcat, parcat, inf_forms)
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
            if oldorth:
                oldorth_inf_forms = generate_forms(generator, lidx, pidx, seg,
                                                   pos, categorisation, OLDORTH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, oldorth_inf_forms,
                                  [TAG_OLDORTH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass
            if ch:
                ch_inf_forms = generate_forms(generator, lidx, pidx, seg,
                                              pos, categorisation, CH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, ch_inf_forms,
                                  [TAG_CH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass

            parcat = Parcat(nonfinite="Inf",
                            function="Cl",
                            tense="Pres")
            categorisation = ["Inf",
                              "Cl"]
            inf_cl_forms = generate_forms(generator, lidx, pidx, seg,
                                          pos, categorisation)
            if not inf_cl_forms:
                inf_cl_forms = [format_inf_cl(inf_form) for inf_form in inf_forms]
            add_forms(formdict, lidx, pidx,
                      lexcat, parcat, inf_cl_forms)
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
            if oldorth:
                oldorth_inf_cl_forms = generate_forms(generator, lidx, pidx, seg,
                                                      pos, categorisation, OLDORTH)
                if not oldorth_inf_cl_forms:
                    oldorth_inf_cl_forms = [format_inf_cl(oldorth_inf_form) for oldorth_inf_form in oldorth_inf_forms]
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, oldorth_inf_cl_forms,
                                  [TAG_OLDORTH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass
            if ch:
                ch_inf_cl_forms = generate_forms(generator, lidx, pidx, seg,
                                                 pos, categorisation, CH)
                if not ch_inf_cl_forms:
                    ch_inf_cl_forms = [format_inf_cl(ch_inf_form) for ch_inf_form in ch_inf_forms]
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, ch_inf_cl_forms,
                                  [TAG_CH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass

            parcat = Parcat(nonfinite="Inf",
                            function="NonCl",
                            tense="Perf")
            add_nonfinite_perf_forms(formdict, lidx, pidx,
                                     lexcat, parcat, [auxiliary], part_perf_forms)
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
            if oldorth:
                add_special_nonfinite_perf_forms(formdict, lidx, pidx,
                                                 lexcat, parcat, [auxiliary], oldorth_part_perf_forms,
                                                 [TAG_OLDORTH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass
            if ch:
                add_special_nonfinite_perf_forms(formdict, lidx, pidx,
                                                 lexcat, parcat, [auxiliary], ch_part_perf_forms,
                                                 [TAG_CH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass

            parcat = Parcat(nonfinite="Inf",
                            function="Cl",
                            tense="Perf")
            add_nonfinite_perf_forms(formdict, lidx, pidx,
                                     lexcat, parcat, [format_inf_cl(auxiliary)], part_perf_forms)
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
            if oldorth:
                add_special_nonfinite_perf_forms(formdict, lidx, pidx,
                                                 lexcat, parcat, [format_inf_cl(auxiliary)], oldorth_part_perf_forms,
                                                 [TAG_OLDORTH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass
            if ch:
                add_special_nonfinite_perf_forms(formdict, lidx, pidx,
                                                 lexcat, parcat, [format_inf_cl(auxiliary)], ch_part_perf_forms,
                                                 [TAG_CH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass

            # participles
            parcat = Parcat(nonfinite="Part",
                            tense="Pres")
            categorisation = ["Part",
                              "Pres"]
            part_pres_forms = generate_forms(generator, lidx, pidx, seg,
                                             pos, categorisation)
            add_forms(formdict, lidx, pidx,
                      lexcat, parcat, part_pres_forms)
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
            if oldorth:
                oldorth_part_pres_forms = generate_forms(generator, lidx, pidx, seg,
                                                         pos, categorisation, OLDORTH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, oldorth_part_pres_forms,
                                  [TAG_OLDORTH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass
            if ch:
                ch_part_pres_forms = generate_forms(generator, lidx, pidx, seg,
                                                    pos, categorisation, CH)
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, ch_part_pres_forms,
                                  [TAG_CH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass

            parcat = Parcat(nonfinite="Part",
                            tense="Perf")
            categorisation = ["Part",
                              "Perf",
                              auxiliary]
            add_forms(formdict, lidx, pidx,
                      lexcat, parcat, part_perf_forms)
            if nonst:
                # no such forms
                pass
            if old:
                # no such forms
                pass
            if oldorth:
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, oldorth_part_perf_forms,
                                  [TAG_OLDORTH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass
            if ch:
                add_special_forms(formdict, lidx, pidx,
                                  lexcat, parcat, ch_part_perf_forms,
                                  [TAG_CH])
                if nonst:
                    # no such forms
                    pass
                if old:
                    # no such forms
                    pass

            # indicative and subjunctive forms
            categorisations = product(TENSES, MOODS, NUMBERS, PERSONS)
            categorisations = filter_categorisations(categorisations, pos)
            for tense, mood, number, person in categorisations:
                parcat = Parcat(person=person,
                                number=number,
                                mood=mood,
                                tense=tense)
                if tense == "Perf":
                    categorisation = [person,
                                      number,
                                      "Pres",
                                      mood]
                    if auxiliary == "haben":
                        forms = generate_forms(generator, LIDX_HABEN, PIDX_HABEN, SEG_HABEN,
                                               pos, categorisation)
                    elif auxiliary == "sein":
                        forms = generate_forms(generator, LIDX_SEIN, PIDX_SEIN, SEG_SEIN,
                                               pos, categorisation)
                    add_perf_forms(formdict, lidx, pidx,
                                   lexcat, parcat, forms, part_perf_forms)
                    if nonst:
                        if auxiliary == "haben":
                            nonst_forms = generate_forms(generator, LIDX_HABEN, PIDX_HABEN, SEG_HABEN,
                                                         pos, categorisation + [NONST])
                        elif auxiliary == "sein":
                            nonst_forms = generate_forms(generator, LIDX_SEIN, PIDX_SEIN, SEG_SEIN,
                                                         pos, categorisation + [NONST])
                        add_special_perf_forms(formdict, lidx, pidx,
                                               lexcat, parcat, nonst_forms, part_perf_forms,
                                               [TAG_NONST])
                    if old:
                        # no such forms
                        pass
                    if oldorth:
                        add_special_perf_forms(formdict, lidx, pidx,
                                               lexcat, parcat, forms, oldorth_part_perf_forms,
                                               [TAG_OLDORTH])
                        if nonst:
                            add_special_perf_forms(formdict, lidx, pidx,
                                                   lexcat, parcat, nonst_forms, oldorth_part_perf_forms,
                                                   [TAG_NONST, TAG_OLDORTH])
                        if old:
                            # no such forms
                            pass
                    if ch:
                        add_special_perf_forms(formdict, lidx, pidx,
                                               lexcat, parcat, forms, ch_part_perf_forms,
                                               [TAG_CH])
                        if nonst:
                            add_special_perf_forms(formdict, lidx, pidx,
                                                   lexcat, parcat, nonst_forms, ch_part_perf_forms,
                                                   [TAG_NONST, TAG_CH])
                        if old:
                            # no such forms
                            pass

                if tense == "PastPerf":
                    categorisation = [person,
                                      number,
                                      "Past",
                                      mood]
                    if auxiliary == "haben":
                        forms = generate_forms(generator, LIDX_HABEN, PIDX_HABEN, SEG_HABEN,
                                               pos, categorisation)
                    elif auxiliary == "sein":
                        forms = generate_forms(generator, LIDX_SEIN, PIDX_SEIN, SEG_SEIN,
                                               pos, categorisation)
                    add_perf_forms(formdict, lidx, pidx,
                                   lexcat, parcat, forms, part_perf_forms)
                    if nonst:
                        # no such forms
                        pass
                    if old:
                        # no such forms
                        pass
                    if oldorth:
                        add_special_perf_forms(formdict, lidx, pidx,
                                               lexcat, parcat, forms, oldorth_part_perf_forms,
                                               [TAG_OLDORTH])
                        if nonst:
                            # no such forms
                            pass
                        if old:
                            # no such forms
                            pass
                    if ch:
                        add_special_perf_forms(formdict, lidx, pidx,
                                               lexcat, parcat, forms, ch_part_perf_forms,
                                               [TAG_CH])
                        if nonst:
                            # no such forms
                            pass
                        if old:
                            # no such forms
                            pass

                elif tense == "Fut":
                    categorisation = [person,
                                      number,
                                      "Pres",
                                      mood]
                    forms = generate_forms(generator, LIDX_WERDEN, PIDX_WERDEN, SEG_WERDEN,
                                           pos, categorisation)
                    add_fut_forms(formdict, lidx, pidx,
                                  lexcat, parcat, forms, inf_forms)
                    if nonst:
                        nonst_forms = generate_forms(generator, LIDX_WERDEN, PIDX_WERDEN, SEG_WERDEN,
                                                     pos, categorisation + [NONST])
                        add_special_fut_forms(formdict, lidx, pidx,
                                              lexcat, parcat, nonst_forms, inf_forms,
                                              [TAG_NONST])
                    if old:
                        # no such forms
                        pass
                    if oldorth:
                        add_special_fut_forms(formdict, lidx, pidx,
                                              lexcat, parcat, forms, oldorth_inf_forms,
                                              [TAG_OLDORTH])
                        if nonst:
                            add_special_fut_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, nonst_forms, oldorth_inf_forms,
                                                  [TAG_NONST, TAG_OLDORTH])
                        if old:
                            # no such forms
                            pass
                    if ch:
                        add_special_fut_forms(formdict, lidx, pidx,
                                              lexcat, parcat, forms, ch_inf_forms,
                                              [TAG_CH])
                        if nonst:
                            add_special_fut_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, nonst_forms, ch_inf_forms,
                                                  [TAG_NONST, TAG_CH])
                        if old:
                            # no such forms
                            pass

                elif tense == "FutPerf":
                    categorisation = [person,
                                      number,
                                      "Pres",
                                      mood]
                    forms = generate_forms(generator, LIDX_WERDEN, PIDX_WERDEN, SEG_WERDEN,
                                           pos, categorisation)
                    add_futperf_forms(formdict, lidx, pidx,
                                      lexcat, parcat, forms, [auxiliary], part_perf_forms)
                    if nonst:
                        nonst_forms = generate_forms(generator, LIDX_WERDEN, PIDX_WERDEN, SEG_WERDEN,
                                                     pos, categorisation + [NONST])
                        add_special_futperf_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, nonst_forms, [auxiliary], part_perf_forms,
                                                  [TAG_NONST])
                    if old:
                        # no such forms
                        pass
                    if oldorth:
                        add_special_futperf_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, forms, [auxiliary], oldorth_part_perf_forms,
                                                  [TAG_OLDORTH])
                        if nonst:
                            add_special_futperf_forms(formdict, lidx, pidx,
                                                      lexcat, parcat, nonst_forms, [auxiliary], oldorth_part_perf_forms,
                                                      [TAG_NONST, TAG_OLDORTH])
                        if old:
                            # no such forms
                            pass
                    if ch:
                        add_special_futperf_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, forms, [auxiliary], ch_part_perf_forms,
                                                  [TAG_CH])
                        if nonst:
                            add_special_futperf_forms(formdict, lidx, pidx,
                                                      lexcat, parcat, nonst_forms, [auxiliary], ch_part_perf_forms,
                                                      [TAG_NONST, TAG_CH])
                        if old:
                            # no such forms
                            pass

                else:
                    categorisation = [person,
                                      number,
                                      tense,
                                      mood]
                    if prev and prev2:
                        forms = generate_forms(generator, lidx, pidx, seg,
                                               pos, categorisation)
                        add_double_prev_verb_forms(formdict, lidx, pidx,
                                                   lexcat, parcat, forms, prev, prev2)
                        if nonst:
                            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                                         pos, categorisation + [NONST])
                            add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                                               lexcat, parcat, nonst_forms, prev, prev2,
                                                               [TAG_NONST])
                        if old:
                            old_forms = generate_forms(generator, lidx, pidx, seg,
                                                       pos, categorisation + [OLD])
                            add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                                               lexcat, parcat, old_forms, prev, prev2,
                                                               [TAG_OLD])
                        if oldorth:
                            oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                           pos, categorisation, OLDORTH)
                            add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                                               lexcat, parcat, oldorth_forms, prev, prev2,
                                                               [TAG_OLDORTH])
                            if nonst:
                                nonst_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                     pos, categorisation + [NONST], OLDORTH)
                                add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                                                   lexcat, parcat, nonst_oldorth_forms, prev, prev2,
                                                                   [TAG_NONST, TAG_OLDORTH])
                            if old:
                                old_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                   pos, categorisation + [OLD], OLDORTH)
                                add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                                                   lexcat, parcat, old_oldorth_forms, prev, prev2,
                                                                   [TAG_OLD, TAG_OLDORTH])
                        if ch:
                            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                      pos, categorisation, CH)
                            add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                                               lexcat, parcat, ch_forms, prev, prev2,
                                                               [TAG_CH])
                            if nonst:
                                nonst_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                                pos, categorisation + [NONST], CH)
                                add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                                                   lexcat, parcat, nonst_ch_forms, prev, prev2,
                                                                   [TAG_NONST, TAG_CH])
                            if old:
                                old_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                              pos, categorisation + [OLD], CH)
                                add_special_double_prev_verb_forms(formdict, lidx, pidx,
                                                                   lexcat, parcat, old_ch_forms, prev, prev2,
                                                                   [TAG_OLD, TAG_CH])

                    elif prev:
                        forms = generate_forms(generator, lidx, pidx, seg,
                                               pos, categorisation)
                        add_prev_verb_forms(formdict, lidx, pidx,
                                            lexcat, parcat, forms, prev)
                        if nonst:
                            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                                         pos, categorisation + [NONST])
                            add_special_prev_verb_forms(formdict, lidx, pidx,
                                                        lexcat, parcat, nonst_forms, prev,
                                                        [TAG_NONST])
                        if old:
                            old_forms = generate_forms(generator, lidx, pidx, seg,
                                                       pos, categorisation + [OLD])
                            add_special_prev_verb_forms(formdict, lidx, pidx,
                                                        lexcat, parcat, old_forms, prev,
                                                        [TAG_OLD])
                        if oldorth:
                            oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                           pos, categorisation, OLDORTH)
                            add_special_prev_verb_forms(formdict, lidx, pidx,
                                                        lexcat, parcat, oldorth_forms, prev,
                                                        [TAG_OLDORTH])
                            if nonst:
                                nonst_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                     pos, categorisation + [NONST], OLDORTH)
                                add_special_prev_verb_forms(formdict, lidx, pidx,
                                                            lexcat, parcat, nonst_oldorth_forms, prev,
                                                            [TAG_NONST, TAG_OLDORTH])
                            if old:
                                old_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                   pos, categorisation + [OLD], OLDORTH)
                                add_special_prev_verb_forms(formdict, lidx, pidx,
                                                            lexcat, parcat, old_oldorth_forms, prev,
                                                            [TAG_OLD, TAG_OLDORTH])
                        if ch:
                            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                      pos, categorisation, CH)
                            add_special_prev_verb_forms(formdict, lidx, pidx,
                                                        lexcat, parcat, ch_forms, prev,
                                                        [TAG_CH])
                            if nonst:
                                nonst_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                                pos, categorisation + [NONST], CH)
                                add_special_prev_verb_forms(formdict, lidx, pidx,
                                                            lexcat, parcat, nonst_ch_forms, prev,
                                                            [TAG_NONST, TAG_CH])
                            if old:
                                old_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                              pos, categorisation + [OLD], CH)
                                add_special_prev_verb_forms(formdict, lidx, pidx,
                                                            lexcat, parcat, old_ch_forms, prev,
                                                            [TAG_OLD, TAG_CH])

                    else:
                        forms = generate_forms(generator, lidx, pidx, seg,
                                               pos, categorisation)
                        add_forms(formdict, lidx, pidx,
                                  lexcat, parcat, forms)
                        if nonst:
                            nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                                         pos, categorisation + [NONST])
                            add_special_forms(formdict, lidx, pidx,
                                              lexcat, parcat, nonst_forms,
                                              [TAG_NONST])
                        if old:
                            old_forms = generate_forms(generator, lidx, pidx, seg,
                                                       pos, categorisation + [OLD])
                            add_special_forms(formdict, lidx, pidx,
                                              lexcat, parcat, old_forms,
                                              [TAG_OLD])
                        if oldorth:
                            oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                           pos, categorisation, OLDORTH)
                            add_special_forms(formdict, lidx, pidx,
                                              lexcat, parcat, oldorth_forms,
                                              [TAG_OLDORTH])
                            if nonst:
                                nonst_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                     pos, categorisation + [NONST], OLDORTH)
                                add_special_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, nonst_oldorth_forms,
                                                  [TAG_NONST, TAG_OLDORTH])
                            if old:
                                old_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                   pos, categorisation + [OLD], OLDORTH)
                                add_special_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, old_oldorth_forms,
                                                  [TAG_OLD, TAG_OLDORTH])
                        if ch:
                            ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                      pos, categorisation, CH)
                            add_special_forms(formdict, lidx, pidx,
                                              lexcat, parcat, ch_forms,
                                              [TAG_CH])
                            if nonst:
                                nonst_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                                pos, categorisation + [NONST], CH)
                                add_special_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, nonst_ch_forms,
                                                  [TAG_NONST, TAG_CH])
                            if old:
                                old_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                              pos, categorisation + [OLD], CH)
                                add_special_forms(formdict, lidx, pidx,
                                                  lexcat, parcat, old_ch_forms,
                                                  [TAG_OLD, TAG_CH])

            # imperative forms
            for number in NUMBERS:
                parcat = Parcat(number=number,
                                mood="Imp")
                categorisation = ["Imp",
                                  number]
                if prev and prev2:
                    forms = generate_forms(generator, lidx, pidx, seg,
                                           pos, categorisation)
                    add_imp_double_prev_verb_forms(formdict, lidx, pidx,
                                                   lexcat, parcat, forms, prev, prev2)
                    if nonst:
                        nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                                     pos, categorisation + [NONST])
                        add_special_imp_double_prev_verb_forms(formdict, lidx, pidx,
                                                               lexcat, parcat, nonst_forms, prev, prev2,
                                                               [TAG_NONST])
                    if old:
                        # no such forms
                        pass
                    if oldorth:
                        oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                       pos, categorisation, OLDORTH)
                        add_special_imp_double_prev_verb_forms(formdict, lidx, pidx,
                                                               lexcat, parcat, oldorth_forms, prev, prev2,
                                                               [TAG_OLDORTH])
                        if nonst:
                            nonst_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                 pos, categorisation + [NONST], OLDORTH)
                            add_special_imp_double_prev_verb_forms(formdict, lidx, pidx,
                                                                   lexcat, parcat, nonst_oldorth_forms, prev, prev2,
                                                                   [TAG_NONST, TAG_OLDORTH])
                        if old:
                            # no such forms
                            pass
                    if ch:
                        ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                  pos, categorisation, CH)
                        add_special_imp_double_prev_verb_forms(formdict, lidx, pidx,
                                                               lexcat, parcat, ch_forms, prev, prev2,
                                                               [TAG_CH])
                        if nonst:
                            nonst_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                            pos, categorisation + [NONST], CH)
                            add_special_imp_double_prev_verb_forms(formdict, lidx, pidx,
                                                                   lexcat, parcat, nonst_ch_forms, prev, prev2,
                                                                   [TAG_NONST, TAG_CH])
                        if old:
                            # no such forms
                            pass

                elif prev:
                    forms = generate_forms(generator, lidx, pidx, seg,
                                           pos, categorisation)
                    add_imp_prev_verb_forms(formdict, lidx, pidx,
                                            lexcat, parcat, forms, prev)
                    if nonst:
                        nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                                     pos, categorisation + [NONST])
                        add_special_imp_prev_verb_forms(formdict, lidx, pidx,
                                                        lexcat, parcat, nonst_forms, prev,
                                                        [TAG_NONST])
                    if old:
                        # no such forms
                        pass
                    if oldorth:
                        oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                       pos, categorisation, OLDORTH)
                        add_special_imp_prev_verb_forms(formdict, lidx, pidx,
                                                        lexcat, parcat, oldorth_forms, prev,
                                                        [TAG_OLDORTH])
                        if nonst:
                            nonst_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                 pos, categorisation + [NONST], OLDORTH)
                            add_special_imp_prev_verb_forms(formdict, lidx, pidx,
                                                            lexcat, parcat, nonst_oldorth_forms, prev,
                                                            [TAG_NONST, TAG_OLDORTH])
                        if old:
                            # no such forms
                            pass
                    if ch:
                        ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                  pos, categorisation, CH)
                        add_special_imp_prev_verb_forms(formdict, lidx, pidx,
                                                        lexcat, parcat, ch_forms, prev,
                                                        [TAG_CH])
                        if nonst:
                            nonst_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                            pos, categorisation + [NONST], CH)
                            add_special_imp_prev_verb_forms(formdict, lidx, pidx,
                                                            lexcat, parcat, nonst_ch_forms, prev,
                                                            [TAG_NONST, TAG_CH])
                        if old:
                            # no such forms
                            pass

                else:
                    forms = generate_forms(generator, lidx, pidx, seg,
                                           pos, categorisation)
                    add_forms(formdict, lidx, pidx,
                              lexcat, parcat, forms)
                    if nonst:
                        nonst_forms = generate_forms(generator, lidx, pidx, seg,
                                                     pos, categorisation + [NONST])
                        add_special_forms(formdict, lidx, pidx,
                                          lexcat, parcat, nonst_forms,
                                          [TAG_NONST])
                    if old:
                        # no such forms
                        pass
                    if oldorth:
                        oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                       pos, categorisation, OLDORTH)
                        add_special_forms(formdict, lidx, pidx,
                                          lexcat, parcat, oldorth_forms,
                                          [TAG_OLDORTH])
                        if nonst:
                            nonst_oldorth_forms = generate_forms(generator, lidx, pidx, seg,
                                                                 pos, categorisation + [NONST], OLDORTH)
                            add_special_forms(formdict, lidx, pidx,
                                              lexcat, parcat, nonst_oldorth_forms,
                                              [TAG_NONST, TAG_OLDORTH])
                        if old:
                            # no such forms
                            pass
                    if ch:
                        ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                  pos, categorisation, CH)
                        add_special_forms(formdict, lidx, pidx,
                                          lexcat, parcat, ch_forms,
                                          [TAG_CH])
                        if nonst:
                            nonst_ch_forms = generate_forms(generator, lidx, pidx, seg,
                                                            pos, categorisation + [NONST], CH)
                            add_special_forms(formdict, lidx, pidx,
                                              lexcat, parcat, nonst_ch_forms,
                                              [TAG_NONST, TAG_CH])
                        if old:
                            # no such forms
                            pass
    return formdict


def get_formdict(generator, lidx, pidx, seg,
                 pos, subcat, person, gender, nonst=False, old=False, oldorth=False, ch=False):
    formdict = defaultdict(list)
    # nouns
    if pos in ["NN", "NPROP"]:
        formdict = get_noun_formdict(generator, lidx, pidx, seg,
                                     pos, gender, nonst, old, oldorth, ch)
    # adjectives
    elif pos == "ADJ":
        formdict = get_adjective_formdict(generator, lidx, pidx, seg,
                                          pos, nonst, old, oldorth, ch)
    # articles
    elif pos == "ART":
        formdict = get_article_formdict(generator, lidx, pidx, seg,
                                        pos, subcat, nonst, old, oldorth, ch)
    # cardinals
    elif pos == "CARD":
        formdict = get_cardinal_formdict(generator, lidx, pidx, seg,
                                         pos, nonst, old, oldorth, ch)
    # ordinals
    elif pos == "ORD":
        formdict = get_ordinal_formdict(generator, lidx, pidx, seg,
                                        pos, nonst, old, oldorth, ch)
    # fractions
    elif pos == "FRAC":
        formdict = get_fraction_formdict(generator, lidx, pidx, seg,
                                         pos, nonst, old, oldorth, ch)
    # demonstrative and possessive pronouns
    elif pos in ["DEM", "POSS"]:
        formdict = get_adjectival_pronoun_formdict(generator, lidx, pidx, seg,
                                                   pos, nonst, old, oldorth, ch)
    # personal pronouns
    elif pos == "PPRO":
        formdict = get_substantival_pronoun_formdict(generator, lidx, pidx, seg,
                                                     pos, subcat, person, gender, nonst, old, oldorth, ch)
    # other pronouns
    elif pos in ["INDEF", "REL", "WPRO"]:
        formdict = get_other_pronoun_formdict(generator, lidx, pidx, seg,
                                              pos, gender, nonst, old, oldorth, ch)
    # verbs
    elif pos == "V":
        formdict = get_verb_formdict(generator, lidx, pidx, seg,
                                     pos, nonst, old, oldorth, ch)
    return formdict


def sort_lemmaspec(lemmaspec):
    lidx = lemmaspec.lidx if lemmaspec.lidx else 0
    pidx = lemmaspec.pidx if lemmaspec.pidx else 0
    seg = lemmaspec.seg_lemma
    pos = POS.index(lemmaspec.pos)
    subcat = SUBCATS.index(lemmaspec.subcat) if lemmaspec.subcat else None
    person = PERSONS.index(lemmaspec.person) if lemmaspec.person else None
    gender = GENDERS.index(lemmaspec.gender) if lemmaspec.gender else None
    return (lidx, pidx, seg,
            pos, subcat, person, gender)


def sort_form(form):
    key = -1
    for index, tags in enumerate([format_tags([TAG_NONST]),
                                  format_tags([TAG_OLD]),
                                  format_tags([TAG_OLDORTH]),
                                  format_tags([TAG_NONST, TAG_OLDORTH]),
                                  format_tags([TAG_OLD, TAG_OLDORTH]),
                                  format_tags([TAG_CH]),
                                  format_tags([TAG_NONST, TAG_CH]),
                                  format_tags([TAG_OLD, TAG_CH])]):
        if form.endswith(tags):
            key = index
    return (key, form)


def create_formdict(generator, analyzer, automaton_type,
                    lemma, lidx=None, pidx=None,
                    pos=None, user_specified=False, nonst=False, old=False, oldorth=False, ch=False):
    lemmaspecs = []
    if user_specified:
        seg = lemma
        if pos == "ART":
            lemmaspecs = [Lemmaspec(lidx, pidx, seg,
                                    pos, subcat, None, None)
                          for subcat in SUBCATS]
        elif pos == "PPRO":
            lemmaspecs = [Lemmaspec(lidx, pidx, seg,
                                    pos, subcat, person, gender)
                          for subcat in SUBCATS
                          for person in PERSONS
                          for gender in GENDERS]
        elif pos in ["INDEF", "NN", "NPROP", "REL", "WPRO"]:
            lemmaspecs = [Lemmaspec(lidx, pidx, seg,
                                    pos, None, None, gender)
                          for gender in GENDERS]
        else:
            lemmaspecs = [Lemmaspec(lidx, pidx, seg,
                                    pos, None, None, None)]
    else:
        traversals = analyze_word(analyzer, lemma, automaton_type)
        lemmaspecs = sorted({Lemmaspec(traversal.lidx, traversal.pidx, seg_lemma(traversal),
                                       traversal.pos, traversal.category, None, None)
                             if traversal.pos == "ART"
                             else Lemmaspec(traversal.lidx, traversal.pidx, seg_lemma(traversal),
                                            traversal.pos, traversal.category, traversal.person, traversal.gender)
                             if traversal.pos == "PPRO"
                             else Lemmaspec(traversal.lidx, traversal.pidx, seg_lemma(traversal),
                                            traversal.pos, None, None, traversal.gender)
                             if traversal.pos in ["INDEF", "NN", "NPROP", "REL", "WPRO"] and traversal.function not in FUNCTIONS
                             else Lemmaspec(traversal.lidx, traversal.pidx, seg_lemma(traversal),
                                            traversal.pos, None, None, None)
                             for traversal in traversals if traversal.analysis == lemma and traversal.pos in POS},
                            key=sort_lemmaspec)
        if lidx in IDX:
            lemmaspecs = [lemmaspec for lemmaspec in lemmaspecs
                          if lemmaspec.lidx == lidx]
        if pidx in IDX:
            lemmaspecs = [lemmaspec for lemmaspec in lemmaspecs
                          if lemmaspec.pidx == pidx]
        if pos in POS:
            lemmaspecs = [lemmaspec for lemmaspec in lemmaspecs
                          if lemmaspec.pos == pos]
    formdict = defaultdict(list)
    for lemmaspec in lemmaspecs:
        formdict_for_lemmaspec = get_formdict(generator, *lemmaspec, nonst=nonst, old=old, oldorth=oldorth, ch=ch)
        for key, value in formdict_for_lemmaspec.items():
            formdict[key] = sorted(set(formdict[key] + value), key=sort_form)

    return formdict


def paradigm_subset(formdict, lidx, pidx, lexcat):
    return [(key, value)
            for key, value in formdict.items()
            if key.lidx == lidx and key.pidx == pidx and key.lexcat == lexcat]


def cat_dict(cats):
    return {key: value
            for key, value in cats._asdict().items()
            if value}


def cat_list(cats):
    # remove duplicates while preserving order
    return {"categories": list(dict.fromkeys(value for value in cat_dict(cats).values()))}


def get_paradigm_dict(lemma, formdict, lidx, pidx, lexcat,
                      no_cats=False, no_lemma=False, empty=False):
    cat_value = cat_list if no_cats else cat_dict
    paradigm_dict = {"lemma": lemma,
                     "lidx": lidx,
                     "pidx": pidx,
                     **cat_value(lexcat),
                     "paradigm": [{**cat_value(key.parcat),
                                   "forms": value}
                                  for key, value in paradigm_subset(formdict, lidx, pidx, lexcat)]}
    if no_lemma:
        for key in ["lemma", "lidx", "pidx"]:
            del paradigm_dict[key]
        for key in list(cat_value(lexcat)):
            del paradigm_dict[key]
    if not empty:
        for key, value in list(paradigm_dict.items()):
            if not value:
                del paradigm_dict[key]
    return paradigm_dict


def get_paradigm_dicts(lemma, formdict,
                       no_cats=False, no_lemma=False, empty=False):
    paradigm_dicts = []
    # remove duplicates while preserving order
    for lidx, pidx, lexcat in list(dict.fromkeys((key.lidx, key.pidx, key.lexcat)
                                                 for key in formdict.keys())):
        paradigm_dict = get_paradigm_dict(lemma, formdict, lidx, pidx, lexcat,
                                          no_cats, no_lemma, empty)
        paradigm_dicts.append(paradigm_dict)
    return paradigm_dicts


def get_value_of_paradigm_key(key, paradigm):
    value = paradigm.get(key, "") or ""
    if isinstance(value, int):
        formatted_value = str(value)
    elif isinstance(value, list):
        if key == "forms":
            formatted_value = ", ".join(value)
        else:
            formatted_value = " ".join(value)
    else:
        formatted_value = value
    return formatted_value


def output_string(string):
    return string


def output_json(paradigms, output_file):
    json.dump(paradigms, output_file, ensure_ascii=False)


def output_yaml(paradigms, output_file):
    yaml.dump(paradigms, output_file, allow_unicode=True, sort_keys=False, explicit_start=True)


def output_dsv(paradigms, output_file, keys, paradigm_keys,
               header=True, plain=False, force_color=False, delimiter="\t"):
    kind = "dumb" if plain else None
    term = Terminal(kind=kind, force_styling=force_color)
    plain = output_string
    csv_writer = csv.writer(output_file, delimiter=delimiter, lineterminator="\n")

    key_format = {"lemma": term.bold_underline,
                  "lidx": term.underline,
                  "pidx": term.underline,
                  "categories": term.underline,
                  "pos": term.underline,
                  "subcat": term.underline,
                  "auxiliary": term.underline,
                  "person": term.underline,
                  "gender": term.underline,
                  "paradigm": {"categories": plain,
                               "degree": plain,
                               "person": plain,
                               "gender": plain,
                               "case": plain,
                               "number": plain,
                               "inflection": plain,
                               "nonfinite": plain,
                               "function": plain,
                               "mood": plain,
                               "tense": plain,
                               "syninfo": plain,
                               "forms": term.bold}}

    if header:
        header_row = [key_format[key](LABEL_MAP[key]) for key in keys]
        header_row += [key_format["paradigm"][key](LABEL_MAP["paradigm"][key]) for key in paradigm_keys]
        csv_writer.writerow(header_row)

    for paradigm_dict in paradigms:
        for paradigm in paradigm_dict["paradigm"]:
            row = [key_format[key](get_value_of_paradigm_key(key, paradigm_dict)) for key in keys]
            row += [key_format["paradigm"][key](get_value_of_paradigm_key(key, paradigm)) for key in paradigm_keys]
            csv_writer.writerow(row)


def output_paradigms(generator, analyzer, output_file, automaton_type,
                     lemma, lidx=None, pidx=None,
                     pos=None, user_specified=False, nonst=False, old=False, oldorth=False, ch=False,
                     no_cats=False, no_lemma=False, empty=False,
                     header=True, plain=False, force_color=False, output_format="tsv"):
    kind = "dumb" if plain else None
    term = Terminal(kind=kind, force_styling=force_color)
    formdict = create_formdict(generator, analyzer, automaton_type,
                               lemma, lidx, pidx,
                               pos, user_specified, nonst, old, oldorth, ch)

    if formdict:
        paradigms = get_paradigm_dicts(lemma, formdict,
                                       no_cats, no_lemma, empty)

        if output_format == "json":
            output_json(paradigms, output_file)
        elif output_format == "yaml":
            output_yaml(paradigms, output_file)
        else:
            keys = [key for key, value in LABEL_MAP.items() if isinstance(value, str)]
            paradigm_keys = [key for key, value in LABEL_MAP["paradigm"].items() if isinstance(value, str)]

            if no_cats:
                for key in LEXCAT:
                    keys.remove(key)
                for key in PARCAT:
                    paradigm_keys.remove(key)
            else:
                keys.remove("categories")
                paradigm_keys.remove("categories")

            if no_lemma:
                keys = []

            if not empty:
                keys_with_values = {key for paradigm_dict in paradigms
                                    for key, value in paradigm_dict.items()
                                    if isinstance(value, (str, int, list))}
                paradigm_keys_with_values = {key for paradigm_dict in paradigms
                                             for paradigm in paradigm_dict["paradigm"]
                                             for key, value in paradigm.items()
                                             if isinstance(value, (str, int, list))}
                keys = [key for key in keys if key in keys_with_values]
                paradigm_keys = [key for key in paradigm_keys if key in key in paradigm_keys_with_values]

            if output_format == "csv":
                output_dsv(paradigms, output_file, keys, paradigm_keys,
                           header, plain, force_color, delimiter=",")
            else:
                output_dsv(paradigms, output_file, keys, paradigm_keys,
                           header, plain, force_color)
    else:
        if lidx and pidx and pos:
            print(f"{progname}: {term.bold(lemma)}: "
                  f"no such lemma of lemma index {lidx}, paradigm index {pidx}, "
                  f"and part-of-speech {pos}",
                  file=sys.stderr)
        elif lidx and pos:
            print(f"{progname}: {term.bold(lemma)}: "
                  f"no such lemma of lemma index {lidx} and part-of-speech {pos}",
                  file=sys.stderr)
        elif pidx and pos:
            print(f"{progname}: {term.bold(lemma)}: "
                  f"no such lemma of paradigm index {pidx} and part-of-speech {pos}",
                  file=sys.stderr)
        elif lidx and pidx:
            print(f"{progname}: {term.bold(lemma)}: "
                  f"no such lemma of lemma index {lidx} and paradigm index {pidx}",
                  file=sys.stderr)
        elif lidx:
            print(f"{progname}: {term.bold(lemma)}: "
                  f"no such lemma of lemma index {lidx}",
                  file=sys.stderr)
        elif pidx:
            print(f"{progname}: {term.bold(lemma)}: "
                  f"no such lemma of paradigm index {pidx}",
                  file=sys.stderr)
        elif pos:
            print(f"{progname}: {term.bold(lemma)}: "
                  f"no such lemma of part-of-speech {pos}",
                  file=sys.stderr)
        elif user_specified:
            print(f"{progname}: {term.bold(lemma)}: "
                  "no user-specified part of speech for lemma",
                  file=sys.stderr)
        else:
            print(f"{progname}: {term.bold(lemma)}: "
                  "no such lemma (for a pseudo-lemma, use --user-specified)",
                  file=sys.stderr)


def main():
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument("lemma",
                            help="lemma (determiners: Fem Nom Sg; nominalised adjectives: Wk)")
        parser.add_argument("output", nargs="?", type=argparse.FileType("w"), default=sys.stdout,
                            help="output file (default: stdout)")
        parser.add_argument("-c", "--csv", action="store_true",
                            help="output CSV table")
        parser.add_argument("-C", "--force-color", action="store_true",
                            help="preserve color and formatting when piping output")
        parser.add_argument("-d", "--automata-location", default=dwdsmor.default_automata_dir,
                            help=f"automata location (default: {format_path(dwdsmor.default_automata_dir)})")
        parser.add_argument("-e", "--empty", action="store_true",
                            help="show empty columns or values")
        parser.add_argument("-H", "--no-header", action="store_false",
                            help="suppress table header")
        parser.add_argument("-i", "--lemma-index", type=int, choices=IDX,
                            help="homographic lemma index")
        parser.add_argument("-I", "--paradigm-index", type=int, choices=IDX,
                            help="paradigm index")
        parser.add_argument("-j", "--json", action="store_true",
                            help="output JSON object")
        parser.add_argument("-n", "--no-cats", action="store_true",
                            help="do not output category names")
        parser.add_argument("-N", "--no-lemma", action="store_true",
                            help="do not output lemma, lemma index, paradigm index, and lexical categories")
        parser.add_argument("-o", "--old", action="store_true",
                            help="output also archaic forms")
        parser.add_argument("-O", "--oldorth", action="store_true",
                            help="output also forms in old spelling")
        parser.add_argument("-p", "--pos", choices=POS,
                            help="part of speech")
        parser.add_argument("-P", "--plain", action="store_true",
                            help="suppress color and formatting")
        parser.add_argument("-s", "--nonst", action="store_true",
                            help="output also non-standard forms")
        parser.add_argument("-S", "--ch", action="store_true",
                            help="output also forms in Swiss spelling")
        parser.add_argument("-t", "--automaton-type", choices=dwdsmor.automaton_types, default="index",
                            help="automaton type (default: index)")
        parser.add_argument("-u", "--user-specified", action="store_true",
                            help="use only user-specified information")
        parser.add_argument("-v", "--version", action="version",
                            version=f"{parser.prog} {version}")
        parser.add_argument("-y", "--yaml", action="store_true",
                            help="output YAML document")
        args = parser.parse_args()

        analyzer = dwdsmor.analyzer(args.automaton_type, Path(args.automata_location))
        generator = dwdsmor.generator(args.automaton_type, Path(args.automata_location))

        if args.json:
            output_format = "json"
        elif args.yaml:
            output_format = "yaml"
        elif args.csv:
            output_format = "csv"
        else:
            output_format = "tsv"

        output_paradigms(generator, analyzer, args.output, args.automaton_type,
                         args.lemma, args.lemma_index, args.paradigm_index,
                         args.pos, args.user_specified, args.nonst, args.old, args.oldorth, args.ch,
                         args.no_cats, args.no_lemma, args.empty,
                         args.no_header, args.plain, args.force_color, output_format)
    except AssertionError as error:
        print(f"{progname}: {error}", file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(130)
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# analysis.py - analyse word forms with DWDSmor
# Andreas Nolda 2026-01-05

import argparse
import csv
import json
import sys
from pathlib import Path

import yaml
from blessings import Terminal

import dwdsmor
from .. import tag


progname = "dwdsmor-analysis"

version = 15.1


LABEL_MAP = {"word": "Wordform",
             "analyses": {"seg_word": "Segmented Wordform",
                          "spec": "Analysis",
                          "lemma": "Lemma",
                          "seg_lemma": "Segmented Lemma",
                          "weights": "Weights",
                          "lidx": "Lemma Index",
                          "pidx": "Paradigm Index",
                          "processes": "Processes",
                          "means": "Means",
                          "pos": "POS",
                          "subcat": "Subcategory",
                          "auxiliary": "Auxiliary",
                          "degree": "Degree",
                          "person": "Person",
                          "gender": "Gender",
                          "case": "Case",
                          "number": "Number",
                          "inflection": "Inflection",
                          "nonfinite": "Nonfinite",
                          "function": "Function",
                          "mood": "Mood",
                          "tense": "Tense",
                          "metainfo": "Metalinguistic",
                          "orthinfo": "Orthography",
                          "syninfo": "Syntax",
                          "ellipinfo": "Ellipsis",
                          "charinfo": "Characters"}}


def analyze_word(analyzer, word, automaton_type=None):
    if automaton_type == "root" or automaton_type == "root2":
        return analyzer.analyze(word, visible_boundaries="+", boundary_tag=" + ", join_tags=True)
    elif automaton_type == "index":
        return analyzer.analyze(word, idx_to_int=True, join_tags=True)
    else:
        return analyzer.analyze(word, join_tags=True)


def analyze_words(analyzer, words, automaton_type=None):
    return tuple(analyze_word(analyzer, word, automaton_type) for word in words)


def generate_words(generator, spec):
    return tuple(traversal.spec for traversal in generator.generate(spec))


def seg_lemma(traversal):
    traversal = traversal.reparse(visible_boundaries=tag.boundary_tags)
    return traversal.analysis


def traversal_asdict(traversal):
    analysis_dict = {"seg_word": None,
                     "spec": traversal.spec,
                     "lemma": traversal.analysis,
                     "seg_lemma": seg_lemma(traversal),
                     "weights": traversal.weights,
                     "lidx": traversal.lidx,
                     "pidx": traversal.pidx,
                     "processes": traversal.processes,
                     "means": traversal.means,
                     "pos": traversal.pos,
                     "subcat": traversal.category,
                     "auxiliary": traversal.auxiliary,
                     "degree": traversal.degree,
                     "person": traversal.person,
                     "gender": traversal.gender,
                     "case": traversal.case,
                     "number": traversal.number,
                     "inflection": traversal.inflection,
                     "nonfinite": traversal.nonfinite,
                     "function": traversal.function,
                     "mood": traversal.mood,
                     "tense": traversal.tense,
                     "metainfo": traversal.metainfo,
                     "orthinfo": traversal.orthinfo,
                     "syninfo": traversal.syninfo,
                     "ellipinfo": traversal.ellipinfo,
                     "charinfo": traversal.charinfo}
    return analysis_dict


def traversals_asdicts(traversals):
    analysis_dicts = [traversal_asdict(traversal) for traversal in traversals]
    return analysis_dicts


def remove_boundaries(seg, boundaries):
    for boundary in boundaries:
        seg = seg.replace(f"<{boundary}>", "")
    return seg


def count_boundaries(seg, boundaries):
    count = sum([seg.count(boundary) for boundary in boundaries])
    return count


def filter_analyses(analyses, keys):
    filtered_analyses = analyses
    for key in keys:
        filtered_analyses = [analysis for analysis in filtered_analyses if analysis[key] is None]
    return filtered_analyses


def get_minimal_analyses(analyses, key, boundaries):
    minimal_analyses = []
    minimum = min([count_boundaries(analysis[key], boundaries) for analysis in analyses], default=-1)
    for analysis in analyses:
        if count_boundaries(analysis[key], boundaries) == minimum:
            minimal_analyses.append(analysis)
    return minimal_analyses


def get_maximal_analyses(analyses, key, boundaries):
    maximal_analyses = []
    maximum = max([count_boundaries(analysis[key], boundaries) for analysis in analyses], default=-1)
    for analysis in analyses:
        if count_boundaries(analysis[key], boundaries) == maximum:
            maximal_analyses.append(analysis)
    return maximal_analyses


def get_minimal_analyses_per_pos(analyses, key, boundaries):
    minimal_analyses = []
    # remove duplicates while preserving order
    for pos in list(dict.fromkeys([analysis["pos"] for analysis in analyses])):
        analyses_with_pos = [analysis for analysis in analyses if analysis["pos"] == pos]
        analyses_with_pos = get_minimal_analyses(analyses_with_pos, key, boundaries)
        minimal_analyses += analyses_with_pos
    return minimal_analyses


def get_maximal_analyses_per_pos(analyses, key, boundaries):
    maximal_analyses = []
    # remove duplicates while preserving order
    for pos in list(dict.fromkeys([analysis["pos"] for analysis in analyses])):
        analyses_with_pos = [analysis for analysis in analyses if analysis["pos"] == pos]
        analyses_with_pos = get_maximal_analyses(analyses_with_pos, key, boundaries)
        maximal_analyses += analyses_with_pos
    return maximal_analyses


def get_matching_seg_words(seg_words, word):
    matching_seg_words = []
    for seg_word in seg_words:
        if remove_boundaries(seg_word, tag.boundary_tags) == word:
            matching_seg_words.append(seg_word)
    return matching_seg_words


def get_unique_analyses(analyses):
    unique_analyses = []
    for analysis in analyses:
        if analysis not in unique_analyses:
            unique_analyses.append(analysis)
    return unique_analyses


def get_value_of_analysis_key(key, analysis):
    value = analysis.get(key, "") or ""
    if isinstance(value, int):
        formatted_value = str(value)
    else:
        formatted_value = value
    return formatted_value


def output_string(string):
    return string


def output_json(word_analyses, output_file):
    json.dump(word_analyses, output_file, ensure_ascii=False)


def output_yaml(word_analyses, output_file):
    yaml.dump(word_analyses, output_file, allow_unicode=True, sort_keys=False, explicit_start=True)


def output_dsv(word_analyses, output_file, keys, analyses_keys,
               header=True, plain=False, force_color=False, delimiter="\t"):
    kind = "dumb" if plain else None
    term = Terminal(kind=kind, force_styling=force_color)
    plain = output_string
    csv_writer = csv.writer(output_file, delimiter=delimiter, lineterminator="\n")

    key_format = {"word": term.bold,
                  "analyses": {"seg_word": term.bright_black,
                               "spec": term.bright_black,
                               "lemma": term.bold_underline,
                               "seg_lemma": term.bright_black_underline,
                               "weights": term.underline,
                               "lidx": term.underline,
                               "pidx": term.underline,
                               "processes": term.underline,
                               "means": term.underline,
                               "pos": term.underline,
                               "subcat": term.underline,
                               "auxiliary": term.underline,
                               "degree": plain,
                               "person": plain,
                               "gender": plain,
                               "case": plain,
                               "number": plain,
                               "inflection": plain,
                               "nonfinite": plain,
                               "function": plain,
                               "mood": plain,
                               "tense": plain,
                               "metainfo": plain,
                               "orthinfo": plain,
                               "syninfo": plain,
                               "ellipinfo": plain,
                               "charinfo": plain}}

    if header:
        header_row = [key_format[key](LABEL_MAP[key]) for key in keys]
        header_row += [key_format["analyses"][key](LABEL_MAP["analyses"][key]) for key in analyses_keys]
        csv_writer.writerow(header_row)

    for word_analyses_dict in word_analyses:
        for analysis in word_analyses_dict["analyses"]:
            row = [key_format[key](get_value_of_analysis_key(key, word_analyses_dict)) for key in keys]
            row += [key_format["analyses"][key](get_value_of_analysis_key(key, analysis)) for key in analyses_keys]
            csv_writer.writerow(row)


def output_analyses(analyzer, generator, input_file, output_file, automaton_type,
                    analysis_string=False, seg_word=False, seg_lemma=False,
                    minimal=False, maximal=False, empty=False, info=True,
                    header=True, plain=False, force_color=False, output_format="tsv"):
    words = tuple(word.strip() for word in input_file.readlines() if word.strip())
    traversals_tuple = analyze_words(analyzer, words, automaton_type)

    if traversals_tuple:
        word_analyses = []
        for word, traversals in zip(words, traversals_tuple):
            analyses = traversals_asdicts(traversals)

            if not info:
                analyses = filter_analyses(analyses, tag.info_tags)

            if minimal:
                analyses = get_minimal_analyses_per_pos(analyses, "seg_lemma", tag.wf_boundary_tags)

            for analysis in analyses:
                if maximal or seg_word:
                    words_tuple = generate_words(generator, analysis["spec"])
                    seg_words = get_matching_seg_words(words_tuple, word)
                    analysis.update({"seg_word": seg_words[0] if seg_words else word})

            if maximal:
                analyses = get_maximal_analyses_per_pos(analyses, "seg_word", tag.infl_boundary_tags)
                analyses = get_maximal_analyses_per_pos(analyses, "seg_word", tag.wf_boundary_tags)

            for analysis in analyses:
                if not seg_word:
                    del analysis["seg_word"]
                if not analysis_string:
                    del analysis["spec"]
                if not seg_lemma:
                    del analysis["seg_lemma"]

                if not empty:
                    for key, value in list(analysis.items()):
                        if not value:
                            del analysis[key]

            analyses = get_unique_analyses(analyses)

            word_analyses.append({"word": word,
                                  "analyses": analyses})

        if output_format == "json":
            output_json(word_analyses, output_file)
        elif output_format == "yaml":
            output_yaml(word_analyses, output_file)
        else:
            keys = [key for key, value in LABEL_MAP.items() if isinstance(value, str)]
            analyses_keys = [key for key, value in LABEL_MAP["analyses"].items() if isinstance(value, str)]

            if not analysis_string:
                analyses_keys.remove("spec")
            if not seg_word:
                analyses_keys.remove("seg_word")
            if not seg_lemma:
                analyses_keys.remove("seg_lemma")

            if not empty:
                keys_with_values = {key for word_analyses_dict in word_analyses
                                    for key, value in word_analyses_dict.items()
                                    if isinstance(value, str)}
                analyses_keys_with_values = {key for word_analyses_dict in word_analyses
                                             for analysis in word_analyses_dict["analyses"]
                                             for key, value in analysis.items()
                                             if isinstance(value, (str, int))}
                keys = [key for key in keys if key in keys_with_values]
                analyses_keys = [key for key in analyses_keys if key in analyses_keys_with_values]

            if output_format == "csv":
                output_dsv(word_analyses, output_file, keys, analyses_keys,
                           header, plain, force_color, delimiter=",")
            else:
                output_dsv(word_analyses, output_file, keys, analyses_keys,
                           header, plain, force_color)


def format_path(path):
    cwd = Path.cwd()
    if path.is_relative_to(cwd):
        return path.relative_to(cwd)
    else:
        return path


def main():
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument("input", nargs="?", type=argparse.FileType("r"), default=sys.stdin,
                            help="input file (one word form per line; default: stdin)")
        parser.add_argument("output", nargs="?", type=argparse.FileType("w"), default=sys.stdout,
                            help="output file (default: stdout)")
        parser.add_argument("-a", "--analysis-string", action="store_true",
                            help="output also analysis string")
        parser.add_argument("-c", "--csv", action="store_true",
                            help="output CSV table")
        parser.add_argument("-C", "--force-color", action="store_true",
                            help="preserve color and formatting when piping output")
        parser.add_argument("-d", "--automata-location", default=dwdsmor.default_automata_dir,
                            help=f"automata location (default: {format_path(dwdsmor.default_automata_dir)})")
        parser.add_argument("-e", "--empty", action="store_true",
                            help="show empty columns or values")
        parser.add_argument("-H", "--no-header", action="store_false",
                            help="suppress table header")
        parser.add_argument("-I", "--no-info", action="store_false",
                            help="do not show analyses with info tags")
        parser.add_argument("-j", "--json", action="store_true",
                            help="output JSON object")
        parser.add_argument("-m", "--minimal", action="store_true",
                            help="prefer lemmas with minimal number of boundaries")
        parser.add_argument("-M", "--maximal", action="store_true",
                            help="prefer word forms with maximal number of boundaries (requires secondary automaton)")
        parser.add_argument("-P", "--plain", action="store_true",
                            help="suppress color and formatting")
        parser.add_argument("-s", "--seg-lemma", action="store_true",
                            help="output also segmented lemma")
        parser.add_argument("-S", "--seg-word", action="store_true",
                            help="output also segmented word form (requires secondary automaton)")
        parser.add_argument("-t", "--automaton-type", choices=dwdsmor.automaton_types, default="lemma",
                            help="type of primary automaton (default: lemma)")
        parser.add_argument("-T", "--automaton2-type", choices=dwdsmor.automaton_types, default="lemma2",
                            help="type of secondary automaton (default: lemma2)")
        parser.add_argument("-v", "--version", action="version",
                            version=f"{parser.prog} {version}")
        parser.add_argument("-y", "--yaml", action="store_true",
                            help="output YAML document")
        args = parser.parse_args()

        analyzer = dwdsmor.analyzer(args.automaton_type, Path(args.automata_location))
        generator = dwdsmor.generator(args.automaton2_type, Path(args.automata_location)) if args.maximal or args.seg_word else None

        if args.json:
            output_format = "json"
        elif args.yaml:
            output_format = "yaml"
        elif args.csv:
            output_format = "csv"
        else:
            output_format = "tsv"

        output_analyses(analyzer, generator, args.input, args.output, args.automaton_type,
                        args.analysis_string, args.seg_word, args.seg_lemma,
                        args.minimal, args.maximal, args.empty, args.no_info,
                        args.no_header, args.plain, args.force_color, output_format)
    except AssertionError as error:
        print(f"{progname}: {error}", file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(130)
    return 0


if __name__ == "__main__":
    sys.exit(main())



from pytest import fixture

import dwdsmor

from .conftest import if_dwds_available


@fixture
def lemmatizer():
    return dwdsmor.lemmatizer()


def test_lemmatizer(lemmatizer):
    assert lemmatizer("getestet", pos={"V"}).analysis == "testen"
    assert lemmatizer("getestet", pos={"ADJ"}).analysis == "getestet"


@if_dwds_available
def test_ptkvz_lemmatization(lemmatizer):
    assert lemmatizer("besser", pos={"V"}, syninfo={"SEP"}).analysis == "besser"
    assert lemmatizer("besser", pos={"ADJ"}).analysis == "gut"



__all__ = [
    "transducer",
    "analyzer",
    "Analyzer",
    "generator",
    "Generator",
    "lemmatizer",
    "Lemmatizer",
    "automaton_types",
    "edition",
    "__version__",
]

import csv
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Union

from sfst_transduce import CompactTransducer, Transducer

from .traversal import Traversal
from .version import __version__

logger = logging.getLogger("dwdsmor")

default_automata_dir = Path(__file__).parent / "automata"
try:
    import dwdsmor_dwds  # type: ignore

    default_automata_dir = Path(dwdsmor_dwds.__file__).parent / "automata"
except ModuleNotFoundError:
    pass

edition_file = default_automata_dir / "EDITION"

edition = "n/a"
if edition_file.is_file():
    edition = edition_file.read_text()


@dataclass
class Generator:
    """
    Wrap a SFST transducer for generating strings, i.e. inflection paradigms.
    """

    transducer: Transducer

    def generate(
        self,
        s: str,
        visible_boundaries="",
        boundary_tag=None,
        join_tags=False,
        idx_to_int=False,
    ) -> Iterable[Traversal]:
        """
        Generate all traversals of a transducer for the given input string.
        """
        return (
            Traversal.parse(
                traversal,
                visible_boundaries=visible_boundaries,
                boundary_tag=boundary_tag,
                join_tags=join_tags,
                idx_to_int=idx_to_int,
            )
            for traversal in self.transducer.generate(s)
        )


@dataclass
class Analyzer:
    """
    Wrap a SFST transducer providing a morphological analysis of input strings.
    """

    transducer: CompactTransducer

    def analyze(
        self,
        s: str,
        visible_boundaries="",
        boundary_tag=None,
        join_tags=False,
        idx_to_int=False,
    ) -> Iterable[Traversal]:
        """
        Generate all traversals of a transducer that express
        morphological analyses of an input string.
        """
        return (
            Traversal.parse(
                traversal,
                visible_boundaries=visible_boundaries,
                boundary_tag=boundary_tag,
                join_tags=join_tags,
                idx_to_int=idx_to_int,
            )
            for traversal in self.transducer.analyse(s)
        )


automaton_types = ("lemma", "lemma2", "finite", "root", "root2", "index")


def assert_valid_automaton_type(automaton_type: str, types=automaton_types):
    assert automaton_type in automaton_types, (
        f"{automaton_type} is not a valid automaton_type. "
        f"Supported types: {automaton_types}"
    )


def transducer(
    automaton_type="lemma",
    automata_dir=default_automata_dir,
    generate=True,
    both_layers=False,
) -> Union[Transducer, CompactTransducer]:
    file_suffix = "a" if generate else "ca"
    transducer_file = automata_dir / f"{automaton_type}.{file_suffix}"
    assert transducer_file.is_file(), f"{transducer_file} does not exist"
    if not generate:
        transducer = CompactTransducer(transducer_file.as_posix())
        transducer.both_layers = both_layers
        return transducer
    else:
        return Transducer(transducer_file.as_posix())


def generator(
    automaton_type="index",
    automata_dir=default_automata_dir,
) -> Generator:
    assert_valid_automaton_type(automaton_type)
    return Generator(transducer(automaton_type, automata_dir))  # type: ignore


def lemma_freqs(automata_dir=default_automata_dir):
    ppm_table = automata_dir / "ppm.csv"
    with ppm_table.open("rt") as ppm_table_f:
        return {lemma: float(ppm) for lemma, ppm in csv.reader(ppm_table_f)}


def analyzer(
    automaton_type="lemma", automata_dir=default_automata_dir, both_layers=False
) -> Analyzer:
    assert_valid_automaton_type(automaton_type)
    return Analyzer(
        transducer(
            automaton_type, automata_dir, generate=False, both_layers=both_layers
        )  # type: ignore
    )


class Lemmatizer:
    def __init__(self, automaton_type="lemma", automata_dir=default_automata_dir):
        self.analyzer = analyzer(automaton_type, automata_dir)
        self.lemma_freqs = lemma_freqs(automata_dir)

    def __call__(self, word, **criteria):
        traversals = tuple(self.analyzer.analyze(word))
        criteria_stack = list((k, v) for k, v in criteria.items() if v)
        criteria_stack.reverse()
        while criteria_stack:
            if len(traversals) == 1:
                break
            attr, attr_vals = criteria_stack.pop()
            filtered = tuple((t for t in traversals if getattr(t, attr) in attr_vals))
            traversals = filtered or traversals
        traversals = sorted(
            traversals, key=lambda t: -self.lemma_freqs.get(t.analysis, 0)
        )
        for traversal in traversals:
            return traversal


def lemmatizer(automaton_type="lemma", automata_dir=default_automata_dir):
    return Lemmatizer(automaton_type, automata_dir)



from itertools import product

stat_tags = ("weights",)
lexeme_tags = ("lidx", "pidx", "pos")
inflection_tags = (
    "category",
    "degree",
    "function",
    "person",
    "gender",
    "case",
    "number",
    "nonfinite",
    "tense",
    "mood",
    "auxiliary",
    "inflection",
)
info_tags = ("metainfo", "orthinfo", "charinfo", "syninfo", "ellipinfo")
wordformation_tags = ("processes", "means")

all_tags = (
    *stat_tags,
    *lexeme_tags,
    *inflection_tags,
    *info_tags,
    *wordformation_tags,
)

tag_values = {
    "": {
        "weights": {"W-"},
        "lidx": {f"IDX{i}" for i in range(1, 9)},
        "pidx": {f"PAR{i}" for i in range(1, 9)},
        "pos": {
            "ADJ",
            "ADV",
            "ART",
            "CARD",
            "CONJ",
            "DEM",
            "FRAC",
            "INDEF",
            "INTJ",
            "NN",
            "NPROP",
            "ORD",
            "POSS",
            "POSTP",
            "PPRO",
            "PREP",
            "PREPART",
            "PROADV",
            "PTCL",
            "PUNCT",
            "REL",
            "V",
            "WPRO",
        },
        "function": {"Attr", "Subst", "UnmFunc"},
        "gender": {"Masc", "Neut", "Fem", "UnmGend"},
        "case": {"Nom", "Acc", "Dat", "Gen", "UnmCase"},
        "number": {"Sg", "Pl", "UnmNum"},
        "inflection": {"St", "Wk", "UnmInfl"},
        "metainfo": {"Old", "NonSt"},
        "orthinfo": {"OLDORTH", "CH"},
        "charinfo": {"CAP"},
        "syninfo": {"MEAS", "SEP"},
        "ellipinfo": {"TRUNC"},
        "processes": {"COMP", "DER", "CONV"},
        "means": {
            "concat",
            "ident",
            "ident|Inf",
            "ident|Part",
            "ident|Fem",
            "ident|Masc",
            "ident|Neut",
            "intf(-)",
            "intf(und)",
            "pref(un)",
            "prev()",
            "prev(ab)",
            "prev(an)",
            "prev(auf)",
            "prev(aus)",
            "prev(bei)",
            "prev(durch)",
            "prev(ein)",
            "prev(fort)",
            "prev(gegen)",
            "prev(heim)",
            "prev(her)",
            "prev(hin)",
            "prev(hinter)",
            "prev(los)",
            "prev(mit)",
            "prev(nach)",
            "prev(ueber)",
            "prev(um)",
            "prev(unter)",
            "prev(vor)",
            "prev(weg)",
            "prev(wieder)",
            "prev(zu)",
            "prev(zurueck)",
            "prev(zwischen)",
            "suff(bar)",
            "suff(e)",
            "suff(er)",
            "suff(chen)",
            "suff(lein)",
            "suff(st)",
            "suff(stel)",
            "suff(zig)",
        },
    },
    "ADJ": {
        "degree": {"Pos", "Comp", "Sup"},
        "function": {"Attr", "Subst", "Attr/Subst", "Pred/Adv", "UnmFunc"},
    },
    "ADV": {"degree": {"Pos", "Comp", "Sup"}},
    "ART": {
        "category": {"Def", "Indef", "Neg"},
    },
    "CONJ": {"category": {"AdjComp", "Coord", "Sub", "InfCl"}},
    "ORD": {
        "function": {"Attr/Subst", "UnmFunc"},
    },
    "POSS": {
        "function": {"Attr", "Subst", "Attr/Subst"},
    },
    "PPRO": {
        "category": {"Pers", "Refl", "Rec"},
        "person": {"1", "2", "3", "UnmPers"},
    },
    "PTCL": {"category": {"AdjSup", "InfCl", "Int", "Neg"}},
    "PUNCT": {
        "category": {
            "Comma",
            "Period",
            "Ellip",
            "Quote",
            "Paren",
            "Dash",
            "Slash",
            "Other",
        },
    },
    "V": {
        "nonfinite": {"Inf", "Part"},
        "function": {"NonCl", "Cl"},
        "person": {"1", "2", "3", "UnmPers"},
        "mood": {"Ind", "Subj", "Imp", "UnmMood"},
        "tense": {"Pres", "Past", "Perf", "UnmTense"},
        "auxiliary": {"haben", "sein"},
    },
}


tag_types = {
    v: tt for _pos, tvs in tag_values.items() for tt, vs in tvs.items() for v in vs
}


aff_infl_boundary = "~"
aff_der_boundary = "-"
intf_boundary = "="
concat_boundary = "#"
prev_boundary = "|"
root_boundary = "+"

infl_boundary_tags = aff_infl_boundary
wf_boundary_tags = (
    aff_der_boundary + intf_boundary + concat_boundary + prev_boundary + root_boundary
)
boundary_tags = infl_boundary_tags + wf_boundary_tags


inflection_tag_seqs = {
    "": (("function", "gender", "case", "number", "inflection"),),
    "ADJ": (("degree", "function", "gender", "case", "number", "inflection"),),
    "ADV": (("degree",),),
    "ART": (("category", "function", "gender", "case", "number", "inflection"),),
    "CONJ": (("category",),),
    "FRAC": (("function",),),
    "INTJ": ((),),
    "NN": (("gender", "case", "number", "inflection"),),
    "NPROP": (("gender", "case", "number", "inflection"),),
    "POSTP": ((),),
    "PPRO": (("category", "person", "gender", "case", "number"),),
    "PREP": ((),),
    "PREPART": (("gender", "case", "number"),),
    "PROADV": ((),),
    "PTCL": (("category"),),
    "V": (
        ("mood", "number"),
        ("nonfinite", "function"),
        ("nonfinite", "tense", "auxiliary"),
        ("person", "number", "tense", "mood"),
    ),
}


def get_taggings(pos, with_info=False):
    tag_seqs = inflection_tag_seqs.get(pos) or inflection_tag_seqs.get("")
    if with_info:
        tag_seqs(ts + info_tags for ts in tag_seqs)

    taggings = set()
    for tag_seq in tag_seqs:
        tagging_values = []
        for tag_type in tag_seq:
            tag_type_values = (
                tag_values.get(pos, dict()).get(tag_type) or tag_values[""][tag_type]
            )
            tagging_values.append(tag_type_values | {None})
        for tagging in product(*tagging_values):
            taggings.add(tuple((t for t in tagging if t is not None)))
    return taggings



import argparse
import csv
import subprocess
from pathlib import Path

from ..tag import all_tags
from ..traversal import Traversal
from ..util import inflected


def traverse(automata_file, sorted=False):
    traversals = subprocess.Popen(
        ["fst-generate", "-b", automata_file.as_posix()],
        encoding=("utf-8" if not sorted else None),
        stdout=subprocess.PIPE,
    )
    if sorted:
        traversals = subprocess.Popen(
            ["sort", "-u"],
            encoding="utf-8",
            stdin=traversals.stdout,
            stdout=subprocess.PIPE,
        )

    with traversals:
        for traversal_str in traversals.stdout:
            yield traversal_str.split()


project_dir = Path(__file__).parent.parent.parent

arg_parser = argparse.ArgumentParser(description="Traverse DWDSmor automaton.")
arg_parser.add_argument(
    "-i",
    "--input-file",
    help="DWDSmor automata to traverse",
    type=Path,
    required=True,
)
arg_parser.add_argument(
    "-o",
    "--output-file",
    help="output CoNLL-U file with (updated) annotations",
    type=argparse.FileType("w"),
    default="-",
)


def main():
    args = arg_parser.parse_args()
    input_file = args.input_file
    output_file = args.output_file

    table = csv.writer(output_file)
    table.writerow(
        [
            "spec",
            "analysis",
            "surface",
            "inflected",
            *all_tags,
        ]
    )
    for surface, spec in traverse(input_file, sorted=True):
        traversal = Traversal.parse(spec)
        table.writerow(
            [
                spec,
                traversal.analysis,
                surface,
                inflected(spec, surface),
                *(getattr(traversal, tt) for tt in all_tags),
            ]
        )


if __name__ == "__main__":
    main()



from io import TextIOWrapper
from pathlib import Path
from tempfile import NamedTemporaryFile
from urllib.request import urlopen
from zipfile import ZipFile

import conllu


def download(dataset, git_ref, paths):
    url = (
        f"https://codeload.github.com/UniversalDependencies/{dataset}/"
        f"zip/refs/tags/{git_ref}"
    )
    path_prefix = f"{dataset}-{git_ref}/"
    temp_zip_file = None
    try:
        with NamedTemporaryFile("wb", delete=False) as tf:
            temp_zip_file = tf.name
            with urlopen(url) as f:
                tf.write(f.read())
        with ZipFile(temp_zip_file) as zf:
            for p in paths:
                p = path_prefix + p
                with zf.open(p, "r") as f, TextIOWrapper(f, encoding="utf-8") as tf:
                    for sentence in conllu.parse_incr(tf):
                        yield sentence
    finally:
        if temp_zip_file is not None:
            Path(temp_zip_file).unlink()


def download_hdt():
    return download(
        "UD_German-HDT",
        "r2.16",
        (f"de_hdt-ud-train-{p1}-{p2}.conllu" for p1 in ("a", "b") for p2 in ("1", "2")),
    )


def download_gsd():
    return download("UD_German-GSD", "r2.17", ("de_gsd-ud-train.conllu",))
